# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius F. Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

from itertools import product
import logging
import re

import numpy as np
import theano
import theano.tensor as T
from theano.tensor.nnet import conv2d
from theano.tensor.nnet.abstract_conv import conv2d_grad_wrt_inputs
from theano.tensor.nnet.conv3d2d import conv3d
from theano.sandbox.cuda import dnn
from theano.tensor.signal.pool import pool_2d

from ..config import config

logger = logging.getLogger('elektronn2log')

dnn_avail = dnn.dnn_available()
logger.warning("Manual dnn calls for conv, are much faster for prediction "
               "and much slower for training.")

dnn_algo = 'small' # Only 'none' is implemented for the conv3d

def apply_except_axis(x, axis, func):
    """
    Apply a contraction function on all but one axis.

    Parameters
    ----------
    x: T.Tensor
        Input tensor.
    axis: int
        Axis to exclude on application.
    func: function
        A function with signature ``func(x, axis=)`` eg T.mean, T.std ...

    Returns
    -------
    T.Tensor
        Contraction of ``x``, but of the same dimensionality.
    """
    x = T.swapaxes(x, 0, axis) # put axis on front
    x = T.flatten(x, 2) # flatten remainder
    y = func(x, axis=1)
    return y


def apply_activation(x, activation_func, b1=None):
    """
    Return an activation function callable matching the name
    Allowed names: 'relu', 'tanh','prelu', 'sigmoid', 'maxout <i>',
    'lin','abs','soft+'.

    Parameters
    ----------
    x: T.Tensor
        Input tensor.
    activation_func: str
        Name of the activation function.
    b1
        Optional b1 parameter for the activation function.
        If this is None, no parameter is passed.

    Returns
    -------
    T.Tensor
        Activation function applied to ``x``.
    """
    func = None
    if activation_func=='tanh': # range = [-1,1]
        func = T.tanh
    elif activation_func=='relu': # rectified linear unit ,range = [0,inf]
        def relu(y):
            return (0.5 * (y + abs(y)))
        func = relu

    elif activation_func=='prelu': # parameterised relu
        def prelu(y, alpha):
            pos = ((y + abs(y)) / 2.0)
            neg = alpha * ((y - abs(y)) / 2.0)
            return pos + neg
        func = prelu

    elif activation_func=='abs': # abs unit ,range = [0,inf]
        func = T.abs_
    elif activation_func in ['sig', 'logistic', 'sigmoid']: # range = [0,1]
        func = T.nnet.sigmoid
    elif activation_func=='soft+':
        func = T.nnet.softplus
    elif activation_func in ['lin', 'linear']:
        def lin(y):
            return y
        func = lin
    elif activation_func=='concentration':
        def func(y):
            #return T.square(T.nnet.softplus(y)) + 2.0
            return T.nnet.softplus(y)*7 + 2.0
    elif activation_func=='radius':
        def func(y):
            return 3+15*T.exp(0.15*y) #T.nnet.softplus(y+5)**1.5
    elif activation_func.startswith("maxout"):
        r = int(re.findall('\d+', activation_func)[0])
        def _maxout(y):
            return maxout(y, factor=r)

        func = _maxout

    else:
        funcs = ['relu', 'tanh','prelu', 'sigmoid', 'maxout <i>',
                 'lin','abs','soft+', 'concentration', 'radius']
        raise NotImplementedError("%s. Permitted activation_funcs :%s" \
                                  %( activation_func, funcs,))

    if b1 is None:
        y = func(x)
    else:
        y = func(x, b1)
    return y


def softmax(x, axis=1, force_builtin=False):
    """
    Calculate softmax (pseudo probabilities).

    Parameters
    ----------
    x: T.Tensor
        Input tensor.
    axis: int
        Axis on which to apply softmax.
    force_builtin: bool
        force usage of ``theano.tensor.nnet.softmax`` (more stable).

    Returns
    -------
    T.Tensor
        ``x`` with softmax applied, same shape.
    """
    if dnn_avail and config.use_manual_cudnn_conv: # oder must always be bc01, and x must always be 4d
        if axis==1:
            dnn_sm = dnn.GpuDnnSoftmax('bc01', 'accurate',  'channel')
            if x.ndim==4:
                logger.debug("Using cuDNN softmax")
                y = dnn_sm(x)
                return y
            elif x.ndim==5: # remap to 4d, this is memory friendly
                logger.debug("Using cuDNN softmax")
                sh = x.shape
                y  = dnn_sm(x.flatten(4)).reshape(sh)
                return y
        # if axis!=1 use own softmax (I don't want to do dimshuffles just to use
        # dnn, it is not that much faster anyway)

    if x.ndim==2 and axis==1:
        return T.nnet.softmax(x)
    elif force_builtin:
        raise NotImplementedError()
    else:
        e_x = T.exp(x - x.max(axis, keepdims=True))
        y   = e_x / e_x.sum(axis, keepdims=True)
        return y

def dot(x, W, axis=1):
    """
    Calculate a tensordot between 1 axis of ``x`` and the first axis of ``W``.

    Requires ``x.shape[axis]==W.shape[0]``.
    Identical to dot if ``x``, ``W`` 2d and ``axis==1``.

    Parameters
    ----------
    x: T.Tensor
        Input tensor.
    W: T.Tensor
        Weight tensor, (f_in, f_out).
    axis: int
        Axis on ``x`` to apply dot.

    Returns
    -------
    T.Tensor
        ``x`` with dot applied. The shape of ``x`` changes on ``axis``
        to ``n_out``.
    """
    if axis<1:
        raise ValueError("Dot on first axis is not supported")

    if x.ndim==2 and axis==1:
        return T.dot(x, W)

    y = T.tensordot(x, W, axes=[axis,0]) # (1, 100, 5, 300, 200) (5, 7) --> (1, 100, 300, 200, 7)
    k = x.ndim - axis
    end = np.roll(np.arange(k)+axis, 1)
    begin = np.arange(axis)
    pattern = np.concatenate((begin,end))
    y = y.transpose(*pattern)
    return y


def upconv(x, w, stride, x_shape=None, w_shape=None, axis_order='dnn'):
    assert stride is not None
    stride = tuple(stride)
    conv_dim = len(stride)
    border_mode = 'valid'

    # if (x_shape is None) or (None in x_shape):  # variable batch size or so
    #     x_shape = None

    if conv_dim==1:
        x = x.dimshuffle(0, 1, 2, 'x')
        w = w.dimshuffle(0, 1, 2, 'x')
        if w_shape is not None:
            w_shape = list(w_shape) + [1, ]
        if x_shape is not None:
            x_shape = list(x_shape) + [1, ]

        stride  = list(stride) + [1, ]
        y = conv2d_grad_wrt_inputs(x, w, x_shape, w_shape, border_mode,
                                   subsample=stride, filter_flip=False)
        y = y[:, :, :, 0]

    elif conv_dim==2:
        y = conv2d_grad_wrt_inputs(x, w, x_shape, w_shape, border_mode,
                                       subsample=stride, filter_flip=False)

    elif conv_dim==3:
        if not dnn_avail or axis_order!='dnn':
            raise ValueError("Need dnn and dnn axis order")
        kerns = dnn.gpu_contiguous(w)
        image = dnn.gpu_contiguous(x)
        k = kerns.shape[1]
        img_sh = list(image.shape)
        out_sh = img_sh[:1] + [k,] + [st*sh for st, sh in zip(stride, img_sh[2:])]
        out = dnn.gpu_alloc_empty(*out_sh)
        desc = dnn.GpuDnnConvDesc(border_mode='valid', subsample=stride,
                                  conv_mode='cross')(out.shape, kerns.shape)
        y = dnn.GpuDnnConv3dGradI()(kerns, image, out, desc)

    return y

### TODO clearer error messages about which arguments can be used
# for dnn and theano respectively. Transform asserts into exceptions
def conv(x, w, axis_order=None, conv_dim=None, x_shape=None, w_shape=None,
         border_mode='valid', stride=None):
    """
    Apply appropriate convolution depending on input and filter dimensionality.
    If input ``w_shape`` is known, conv might be replaced by tensordot

    There are static assumptions which axes are spatial.

    Parameters
    ----------
    x: T.Tensor
        | Input data (mini-batch).
        | Tensor of shape ``(b, f, x)``, ``(b, f, x, y)``, ``(b, z, f, x, y)``
          or ``(b,f,x,y,z)``.
    w: T.Tensor
        | Set of convolution filter weights.
        | Tensor of shape ``(f_out, f_in, x)``, ``(f_out, f_in, x, y)``,
          ``(f_out, z, f_in, x, y)`` or ``(f_out, f_in, x, y, z)``.
    axis_order: str
        | (only relevant for 3d)
        | ``'dnn'`` ``(b,f,x,y(,z))`` or ``'theano'`` ``(b, z, f, x, y)``.
    conv_dim: int
        Dimensionality of the applied convolution (not the absolute dim of
        the inputs).
    x_shape: tuple
        shape tuple (``TaggedShape`` supported).
    w_shape: tuple
        shape tuple, see ``w``.
    border_mode: str
        * ``'valid'``: only apply filter to complete patches of the image.
          Generates output of shape: image_shape -filter_shape + 1.
        * ``'full'`` zero-pads image to multiple of filter shape to generate
          output of shape: image_shape + filter_shape - 1.
    stride: tuple
        | (tuple of len 2)
        | Factor by which to subsample the output.

    Returns
    -------
    T.Tensor
        Set of feature maps generated by convolution.
    """
    if (x_shape is None) or (None in x_shape): # variable batch size or so
        x_shape = None

    assert axis_order in ['dnn', 'theano', None]

    if conv_dim is not None:
        if x.ndim!=conv_dim+2 or w.ndim!=conv_dim+2:
            raise ValueError("Cannot perform %id conv on input and filter of"
                             "dim %i, %i" % (conv_dim, x.ndim, w.ndim))

    else: # infer conv_dim
        conv_dim = x.ndim-2
        if w.ndim!=conv_dim+2:
            raise ValueError("Dimension mismatch for conv: tried to do %id conv"
                             "on %id input x. This requires %id filter, but got"
                             "%id" % (conv_dim, x.ndim, x.ndim, w.ndim))
        if conv_dim>3:
            raise ValueError("Input tensor dim to big. No conv for dim>5.")

    if border_mode=='same':
        assert w_shape is not None
        assert np.all(np.remainder(w_shape[-conv_dim:], 2)==1), "For conv same kernels must be uneven"
        border_mode='full'
        crop_full = True
    else:
        crop_full = False

    use_tensordot = False
    if (w_shape is not None) and (stride is None): # cannot use tensordot with strides
        if conv_dim<3 or axis_order=='dnn':
            use_tensordot = np.all(np.equal(w_shape[2:], 1))
        else: # theano order for 3d conv
            use_tensordot = w_shape[1] == 1 and np.all(np.equal(w_shape[3:], 1))

    y = None
    if conv_dim==1:
        x = x.dimshuffle(0, 1, 2, 'x')
        w = w.dimshuffle(0, 1, 2, 'x')
        if w_shape is not None:
            w_shape = list(w_shape) + [1, ]
        if x_shape is not None:
            x_shape = list(x_shape) + [1,]
        if stride is None:
            stride = (1, 1)
        y = conv2d(x, w, x_shape, w_shape, border_mode, subsample=stride)
        y = y[:, :, :, 0]

    elif conv_dim==2:
        if stride is None:
            stride = (1, 1)
        if use_tensordot:
            logger.debug("Using dot for 2d conv")
            w = w[:, :, 0, 0].T # (f_in, f_out) (5, 7)
            y = dot(x, w, axis=1)

        elif dnn_avail and config.use_manual_cudnn_conv:
            logger.debug("Using cuDNN 2dconv")
            y = dnn.dnn_conv(x, w, border_mode, subsample=stride, algo=dnn_algo)
        else: # fallback to theano
            y = conv2d(x, w, x_shape, w_shape, border_mode, subsample=stride)

    elif conv_dim==3:
        assert axis_order in ['dnn', 'theano']
        use_dnn = dnn_avail
        if not config.use_manual_cudnn_conv:
            use_dnn = False
        if w_shape[2]==1 and config.use_manual_cudnn_conv_not_w1:
            use_dnn = False
            logger.debug("Ignoring manual 3d cuDNN conv because kernel is "
            "1 for first axis") # then theano automatically uses dnn 2d conv which
            # has faster gradient than dnn 3d conv

        if stride is not None:
            raise NotImplementedError("Cannot use strided conv with 3d conv")
        if use_tensordot:
            logger.debug("Using dot for 3d conv")
            if axis_order=='theano':
                w = w[:, 0, :, 0, 0].T # (f_in, f_out)
                y = dot(x, w, axis=2)
            elif axis_order=='dnn':
                w = w[:, :, 0, 0, 0].T # (f_in, f_out)
                y = dot(x, w, axis=1)

        elif use_dnn:
            if stride is None:
                stride = (1, 1, 1)
            if axis_order=='dnn':
                logger.debug("Using cuDNN 3dconv")
                y = dnn.dnn_conv3d(x, w, border_mode, subsample=stride, algo=dnn_algo) # (b, f, x, y, z)
            else:
                if config.show_axis_order_warning:
                    logger.warning("cuDNN available but axis order is "
                    "for theano (z before f). This leads to possibly "
                    "inefficient dimshuffles. use cuDNN axis order.\n"
                    "Using dnn 3dconv")
                x = x.dimshuffle(0,2,1,3,4)
                w = w.dimshuffle(0, 2, 1, 3, 4)
                y = dnn.dnn_conv3d(x, w, border_mode, subsample=stride, algo=dnn_algo) # (b, f, x, y, z)
                y = y.dimshuffle(0,2,1,3,4)

        else: # fallback to theano
            if axis_order=='theano':
                logger.debug("Using theano 3dconv")
                y = conv3d(x, w, x_shape, w_shape, border_mode) # (b, z, f, x, y)
            else:
                if config.use_manual_cudnn_conv and not dnn_avail:
                    if config.show_axis_order_warning:
                        logger.warning("cuDNN not available but axis order is"
                         "for cuDNN (z after features). This leads to possibly "
                         "inefficient dimshuffles Use theano axis order or "
                         "install cuDNN.\nUsing theano 3dconv")
                x = x.dimshuffle(0,2,1,3,4)
                w = w.dimshuffle(0, 2, 1, 3, 4)
                # Also swap shapes!
                w_shape = list(w_shape)
                z,f = w_shape[1], w_shape[2]
                w_shape[2] = z
                w_shape[1] = f
                if x_shape is not None:
                    x_shape = list(x_shape)
                    z,f = x_shape[1], x_shape[2]
                    x_shape[2] = z
                    x_shape[1] = f

                y = conv3d(x, w, x_shape, w_shape, border_mode) # (b, z, f, x, y)
                y = y.dimshuffle(0,2,1,3,4)

    if crop_full:
        cropper = []
        off = np.divide(w_shape[-conv_dim:], 2).astype(np.int)
        k = 0
        if axis_order=='theano' and conv_dim==3:
            for i in range(y.ndim):
                if i in [1,3,4]:
                    cropper.append(slice(off[k], -off[k]))
                    k += 1
                else:
                    cropper.append(slice(None))
        else:
            for i in range(y.ndim):
                if i >= y.ndim - conv_dim:
                    cropper.append(slice(off[k], -off[k]))
                    k += 1
                else:
                    cropper.append(slice(None))

        cropper = tuple(cropper)
        y = y[cropper]

    return y


def maxout(x, factor=2, axis=None):
    """
    Maxpooling along the feature axis.

    The feature count is reduces by ``factor``.

    Parameters
    ----------
    x: T.Tensor
        Input tensor (b, f, x, y), (b, z, f, x, y).
    factor: int
        Pooling factor.
    axis: int or None
        Feature axis of ``x`` (1 or 2).
        If None, 5d tensors get axis 2 and all others axis 1.

    Returns
    -------
    T.Tensor
        ``x`` with pooling applied.
    """
    if axis is None:
        axis = 2 if x.ndim==5 else 2

    if axis not in [1,2]:
        raise ValueError("Maxout only permitted on axis 1 or 2")

    y = None
    if axis==1:
        y =  x[:,0::factor]
        for i in range(1, factor):
            t = x[:,i::factor]
            y = T.maximum(y, t)

    elif axis==2:
        y =  x[:,:,0::factor]
        for i in range(1, factor):
            t = x[:,:,i::factor]
            y = T.maximum(y, t)

    return y


# def pooling_3d2d(x, pool, spatial_axes, func=T.max,):
#     """
#     Pooling along spatial axes of 3d and 2d tensors.
#     There are static assumptions which axes are spatial.
#     The axes of ``x`` must be divisible by the corresponding pooling factor,
#     otherwise the computation might crash later.
#     :param x: tensor (b, f, x, y), (b, z, f, x, y)
#     :param pool:2/3-tuple of pooling factors. They refer to the spatial
#     axes of ``x`` (x,y)/(z,x,y)
#     :param func: function with signature ``f(x, axis=)`` e.g. ``T.mean``, ``T.max``
#     :return: ``x`` with pooling applied. The spatial axes are decrease by the
#     corresponding pooling factors
#     """
#     spatial_axes = list(spatial_axes)
#     _pool = np.ones(x.ndim, np.int)
#     _pool[spatial_axes] = pool
#     return pooling_nd(x, _pool, func)
#
#
# def pooling_nd(x, pool, func=T.max):
#     """
#     Pooling along spatial axes of nd tensors.
#     The axes of ``x`` must be divisible by the corresponding pooling factor,
#     otherwise the computation might crash later.
#     :param x: nd tensor
#     :param pool: tuple of length n
#     :param func: function with signature ``f(x, axis=)`` e.g. ``T.mean``, ``T.max``
#     :return: ``x`` with pooling applied. The axis lengths are decrease by the
#     corresponding pooling factors
#     """
#     accum =[]
#     for i,ix in enumerate(product(*map(lambda x: list(np.arange(x)), pool))):
#         sl = tuple([slice(x[0], x[1], x[2]) for x in zip(ix, x.shape, pool)])
#         accum.append(x[sl])
#
#     accum = T.stacklists(accum)
#     y = func(accum, axis=0)
#     return y


def pooling(x, pool, spatial_axes, mode='max', stride=None):
    """
    Pooling along spatial axes of 3d and 2d tensors.

    There are static assumptions which axes are spatial.
    The spatial axes must be divisible by the corresponding pooling factor,
    otherwise the computation might crash later.

    Parameters
    ----------
    x: T.Tensor
        Input tensor (b, f, x, y), (b, z, f, x, y).
    pool: tuple
        2/3-tuple of pooling factors.
        They refer to the spatial axes of ``x`` (x,y)/(z,x,y).
    spatial_axes: tuple
    mode: str
        Can be any of the modes supported by Theano's dnn_pool():
        ('max', 'average_inc_pad', 'average_exc_pad', 'sum').

        'max' (default): max-pooling
        'average' or 'average_inc_pad': average-pooling
        'sum': sum-pooling
    stride: tuple

    Returns
    -------
    T.Tensor
        ``x`` with maxpooling applied. The spatial axes are decreased by the
        corresponding pooling factors
    """  # TODO: Some params are undocumented.
    if np.all(np.equal(pool, 1)): # Short circuit no pool
        return x

    pool = tuple(pool)
    if stride is None:
        stride = pool
    else:
        stride = tuple(stride)

    spatial_axes = list(spatial_axes)
    if not dnn_avail and config.use_manual_cudnn_pool and mode!='max':
        logger.warning("Pooling modes can only be selected if cuDNN is available, mode is ignored")

    if spatial_axes==[2,3,4]:
        axis_order = 'dnn'
    elif spatial_axes==[1,3,4]:
        axis_order = 'theano'
    else:
        axis_order = None

    if mode == 'average':
        mode = 'average_inc_pad'  # Theano's internal name. 'average' is deprecated.

    ndim = len(pool)
    if ndim==3:
        if not axis_order:
            raise ValueError("Axis order not recognised, must be [2,3,4] (dnn) or [1,3,4] (theano).")
        if dnn_avail and config.use_manual_cudnn_pool:
            pad = (0,0,0)
            if axis_order=='dnn':
                logger.debug("Using dnn 3dpool")
                y = dnn.dnn_pool(x, pool, stride=stride, pad=pad, mode=mode) # (b, f, x, y, z)
            else:
                logger.warning("cuDNN is available but the used axis order is "
                "for theano (z before features). This requires possibly "
                "inefficient dimshuffles, consider using cuDNN axis order. "
                "Using dnn 3pool")

                x = x.dimshuffle(0,2,1,3,4)
                y = dnn.dnn_pool(x, pool, stride=stride, pad=pad, mode=mode) # (b, f, x, y, z)
                y = y.dimshuffle(0,2,1,3,4)

        else: # fallback to theano
            if pool != stride:
                raise NotImplementedError("Stride!=Pool using theano 3d pooling (dnn pooling will work)")

            if axis_order=='theano':
                logger.debug("Using theano 3dpool")
                y = pool_2d(x, pool[1:], st=stride[1:], ignore_border=True) # (b, z, f, x, y)
                m = y[:,0::pool[0]]
                for z in range(1,pool[0]): ### TODO obey stride
                    t = y[:, z::pool[0]]
                    m = T.maximum(t, m) # (b, z, f, x, y)
                y = m
            else:
                logger.debug("Using theano 3dpool")
                y = pool_2d(x, pool[1:], st=stride[1:], ignore_border=True) # (b, z, f, x, y)
                m = y[:, :, 0::pool[0]]
                for z in range(1,pool[0]): ### TODO obey stride
                    t = y[:, :, z::pool[0]]
                    m = T.maximum(t, m)

                y = m

    elif ndim==2:
        if spatial_axes!=[2,3]:
            raise NotImplemented("Can only pool on last axes [2,3], this input hat spatial axes %s" %(spatial_axes,))
        if dnn_avail and config.use_manual_cudnn_pool:
            y = dnn.dnn_pool(x, pool, stride=stride, pad=(0,0), mode=mode)
        else:
            y = pool_2d(x, pool, st=stride, ignore_border=True)

    elif ndim==1:
        x = x.dimshuffle(0,1,2,'x')
        pool = [pool[0], 1]
        stride = [stride[0], 1]
        y = pool_2d(x, pool, st=stride ,ignore_border=True)[:,:,:,0]
    else:
        raise NotImplementedError("Only 1/2/3-dim maxpooling with this function.")

    return y


def fragmentpool(conv_out, pool, offsets, strides, spatial_axes, mode='max'):
    if np.all(np.equal(pool, 1)): # Short circuit no pool
        return conv_out, offsets, strides

    spatial_axes = list(spatial_axes)
    result = []
    offsets_new = []
    offsets = np.array(offsets, np.int)
    strides = np.array(strides, np.int)
    sh = conv_out.shape
    _pool = np.ones(conv_out.ndim, np.int)
    _pool[spatial_axes] = pool
    ###TODO maybe unify this loop and the pooling loop for speedup... timing!
    for i,ix in enumerate(product(*map(lambda x: list(np.arange(x)), _pool))):
        sl = tuple([slice(x[0], x[0] + x[1] - x[2] + 1) for x in zip(ix, sh, _pool)])
        result.append(pooling(conv_out[sl], pool, spatial_axes, mode=mode))
        for p in offsets:
            new = p.copy()
            ix_spatial = [ix[ax] for ax in spatial_axes]
            new += np.multiply(ix_spatial, strides)
            offsets_new.append(new)

    result = T.concatenate(result, axis=0)
    offsets_new = np.array(offsets_new)
    strides_new = np.multiply(pool, strides)

    return result, offsets_new, strides_new


def fragments2dense(fragments, offsets, strides, spatial_axes):
    spatial_axes = list(spatial_axes)
    example_stride = np.prod(strides) # This strides is conceptually unneeded but theano-grad fails otherwise
    sh             = fragments.shape
    #spatial_axes   = [1,3,4] if len(strides)==3 else [2,3]
    out_sh         = list(sh)
    out_sh[0]      = 1
    for i,ax in enumerate(spatial_axes):
        out_sh[ax] *= strides[i]

    # assert np.prod(out_sh)==np.prod(sh) # cannot work on symbolical....
    zero = np.array((0,), dtype=theano.config.floatX)
    embedding = T.alloc(zero, *out_sh)
    for i,off in enumerate(offsets):
        sl = [slice(None),]*len(out_sh) # defaults to ":" slice
        for k,ax in enumerate(spatial_axes):
            sl[ax] = slice(off[k], None, strides[k])
        embedding = T.set_subtensor(embedding[tuple(sl)],
                                    fragments[i::example_stride])

    return embedding


def upsampling_nd(x, pool):
    new_sh = [xi*pi for xi, pi in zip(x.shape, pool)]
    zero = T.cast(0, x.dtype)
    out = T.alloc(zero, *new_sh) # T.zeros is slower
    for i,ix in enumerate(product(*map(lambda s: list(np.arange(s)), pool))):
        sl = tuple([slice(s[0], s[1], s[2]) for s in zip(ix, out.shape, pool)])
        out = T.set_subtensor(out[sl], x)

    return out


def upsampling(x, pool, spatial_axes):
    """
    Upsamling through repetition: s_new = s*p.

    e.g for pool=3: aaabbbccc...

    Parameters
    ----------
    x: T.Tensor
        Input tensor.
    pool: int
        Upsampling factor.
    spatial_axes: list
        List of axes on which to perform upsampling.

    Returns
    -------
    T.Tensor
        ``x`` with upsampling applied.
    """
    """
    Upsamling through repetition: s_new = s*p
    e.g for p=3: aaabbbccc...
    :param x:
    :param pool:
    :param spatial_axes:
    :return:
    """
    spatial_axes = list(spatial_axes)
    _pool = np.ones(x.ndim, np.int)
    _pool[spatial_axes] = pool
    return upsampling_nd(x, _pool)


def unpooling_nd(x, pool):
    new_sh = [xi*pi + (pi-1) for xi, pi in zip(x.shape, pool)]
    zero = T.cast(0, x.dtype)
    out = T.alloc(zero, *new_sh) # T.zeros is slower
    sl = tuple([slice(pi-1, xi*pi, pi) for xi, pi in zip(x.shape, pool)])
    out = T.set_subtensor(out[sl], x)

    return out


def unpooling(x, pool, spatial_axes):
    """
    Symmetric unpooling with border: s_new = s*pool + pool-1.

    Insert values strided, e.g for pool=3: 00x00x00x...00x00.

    Parameters
    ----------
    x: T.Tensor
        Input tensor.
    pool: int
        Unpooling factor.
    spatial_axes: list
        List of axes on which to perform unpooling.

    Returns
    -------
    T.Tensor
        ``x`` with unpooling applied.
    """
    spatial_axes = list(spatial_axes)
    _pool = np.ones(x.ndim, np.int)
    _pool[spatial_axes] = pool
    return unpooling_nd(x, _pool)



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

import logging

import numpy as np
import theano.tensor as T

from .computations import softmax
from .graphutils import TaggedShape, floatX
from .node_basic import Node, FromTensor
from .variables import VariableParam
from .neural import Conv

logger = logging.getLogger('elektronn2log')
inspection_logger = logging.getLogger('elektronn2log-inspection')

__all__ = ['GaussianNLL', 'BinaryNLL', 'AggregateLoss', 'SquaredLoss',
           'AbsLoss',
           'Softmax', 'MultinoulliNLL', 'MalisNLL', 'Errors', 'BetaNLL',
           'SobelizedLoss', 'BlockedMultinoulliNLL', 'OneHot',
           'EuclideanDistance', 'RampLoss']

xlogy0 = T.xlogx.xlogy0
EPS = 1e-5


class Softmax(Node):
    """
    Softmax node.

    Parameters
    ----------
    parent: Node
        Input node.
    n_class
    n_indep
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, parent, n_class='auto', n_indep=1, name="softmax",
                 print_repr=True):

        super(Softmax, self).__init__(parent, name, print_repr)

        n_f = parent.shape['f']

        if hasattr(parent, 'activation_func'):
            if parent.activation_func != 'lin':
                raise ValueError("The parent of a Softmax-node must have a "
                                 "linear activation function.")

        if n_class == 'auto':
            if n_f % n_indep == 0:
                n_class = n_f // n_indep
            else:
                n_class = n_f // n_indep
                raise ValueError("Cannot create %i-fold %i-class softmax "
                                 "from %i features." % (n_indep, n_class, n_f))
        else:
            if n_class * n_indep != n_f:
                raise ValueError("Cannot create %i-fold %i-class softmax ")

        self.n_class = n_class
        self.n_indep = n_indep

    def _make_output(self):
        """ Computation of Theano Output """
        n_class = self.n_class
        n_indep = self.n_indep
        x = self.parent.output
        axis = self.parent.shape.tag2index('f')

        if self.n_indep == 1:
            self.output = softmax(x, axis=axis)
        else:
            y = []
            for i in range(n_indep):
                sl = [slice(None), ] * x.ndim
                sl[axis] = slice(i * n_class, (i + 1) * n_class, 1)
                y_part = softmax(x[tuple(sl)], axis=axis)
                y.append(y_part)

            y = T.concatenate(y, axis=axis)
            self.output = y


class OneHot(Node):
    """
    Onehot node.

    Parameters
    ----------
    target: T.Tensor
        Target tensor.
    n_class: int
    axis
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, target, n_class, axis='f', name="onehot",
                 print_repr=True):
        super(OneHot, self).__init__(target, name, print_repr)

        self.target = target
        self.axis = target.shape.tag2index('f')
        self.n_class = n_class

    def _make_output(self):
        """ Computation of Theano Output """
        target = self.target.output

        pattern_exp_class = ['x', ] * target.ndim
        pattern_exp_class[self.axis] = 0

        classes = T.arange(self.n_class)
        classes = classes.dimshuffle(pattern_exp_class)

        target = T.addbroadcast(target, self.axis)
        target = T.eq(target, classes)  # to 1-hot
        target = T.cast(target, floatX)

        self.output = target

    def _calc_shape(self):
        sh = self.parent.shape.updateshape(self.axis, self.n_class)
        self.shape = sh


class MultinoulliNLL(Node):
    """
    Returns the symbolic mean and instance-wise negative log-likelihood of the prediction
    of this model under a given target distribution.

    Parameters
    ----------
    pred: Node
        Prediction node.
    target: T.Tensor
        corresponds to a vector that gives the correct label for each example. Labels < 0 are ignored (e.g. can
        be used for label propagation).
    target_is_sparse: bool
        If the target is sparse.
    class_weights: T.Tensor
        weight vector of float32 of length  ``n_lab``. Values: ``1.0`` (default), ``w < 1.0`` (less important),
        ``w > 1.0`` (more important class).
    example_weights: T.Tensor
        weight vector of float32 of shape ``(bs, z, x, y)`` that can give the individual examples (i.e. labels for
        output pixels) different weights. Values: ``1.0`` (default), ``w < 1.0`` (less important),
        ``w > 1.0`` (more important example). Note: if this is not normalised/bounded it may result in a
        effectively modified learning rate!

    The following refers to lazy labels, the masks are always on a per patch basis, depending on the
    origin cube of the patch. The masks are properties of the individual image cubes and must be loaded
    into CNNData.

    mask_class_labeled: T.Tensor
        shape = (batchsize, num_classes).
        Binary masks indicating whether a class is properly labeled in ``y``. If a class ``k``
        is (in general) present in the image patches **and** ``mask_class_labeled[k]==1``, then
        the labels  **must** obey ``y==k`` for all pixels where the class is present.
        If a class ``k`` is present in the image, but was not labeled (-> cheaper labels), set
        ``mask_class_labeled[k]=0``. Then all pixels for which the ``y==k`` will be ignored.
        Alternative: set ``y=-1`` to ignore those pixels.
        Limit case: ``mask_class_labeled[:]==1`` will result in the ordinary NLL.
    mask_class_not_present: T.Tensor
        shape = (batchsize, num_classes).
        Binary mask indicating whether a class is present in the image patches.
        ``mask_class_not_present[k]==1`` means that the image does **not** contain examples of class ``k``.
        Then for all pixels in the patch, class ``k`` predictive probabilities are trained towards ``0``.
        Limit case: ``mask_class_not_present[:]==0`` will result in the ordinary NLL.
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.

    Examples
    --------

    - A cube contains no class ``k``. Instead of labelling the remaining classes they can be
      marked as unlabelled by the first mask (``mask_class_labeled[:]==0``, whether ``mask_class_labeled[k]``
      is ``0`` or ``1`` is actually indifferent because the labels should not be ``y==k`` anyway in this case).
      Additionally ``mask_class_not_present[k]==1`` (otherwise ``0``) to suppress predictions of ``k`` in
      in this patch. The actual value of the labels is indifferent, it can either be ``-1`` or it could be the
      background class, if the background is marked as unlabelled (i.e. then those labels are ignored).

    - Only part of the cube is densely labelled. Set ``mask_class_labeled[:]=1`` for all classes, but set the
      label values in the unlabelled part to ``-1`` to ignore this part.

    - Only a particular class ``k`` is labelled in the cube. Either set all other label pixels to ``-1`` or the
      corresponding flags in ``mask_class_labeled`` for the unlabelled classes.

    ..  Note::
        Using ``-1`` labels or telling that a class is not labelled, is somewhat redundant and just
        supported for convenience.
    """
    # TODO: add comment on normalisation.

    def __init__(self, pred, target, target_is_sparse=False, class_weights=None,
                 example_weights=None, mask_class_labeled=None,
                 mask_class_not_present=None, name="nll", print_repr=True):
        parents = [pred, target]
        if class_weights is not None:
            if isinstance(class_weights, Node):
                parents.append(class_weights)
            else:
                class_weights = np.array(class_weights, dtype=floatX)
                class_weights = VariableParam(value=class_weights,
                                              name="class_weights",
                                              dtype=floatX,
                                              apply_train=False)
        if example_weights is not None:
            parents.append(example_weights)
        if mask_class_labeled is not None:
            parents.append(mask_class_labeled)
        if mask_class_not_present is not None:
            parents.append(mask_class_not_present)

        super(MultinoulliNLL, self).__init__(parents, name, print_repr)

        if isinstance(pred, Softmax):
            parent = pred
        else:
            if isinstance(pred, FromTensor) and isinstance(pred.parent,
                                                           Softmax):
                parent = pred.parent  # splitted softmax...
            else:
                raise ValueError(
                    "The prob input to a MultinoulliNLL-node must be "
                    "a Softmax-Node.")

        self.target = target
        self.pred = pred
        self.axis = pred.shape.tag2index('f')
        self.n_class = parent.n_class
        self.n_indep = parent.n_indep
        self.target_is_sparse = target_is_sparse
        self.class_weights = class_weights
        self.example_weights = example_weights
        self.mask_class_labeled = mask_class_labeled
        self.mask_class_not_present = mask_class_not_present

    def _make_output(self):
        """ Computation of Theano Output """
        pred = self.pred.output
        target = self.target.output

        pattern_add_class = list(range(pred.ndim - 1))
        pattern_add_class.insert(self.axis, 'x')

        pattern_exp_class = ['x', ] * pred.ndim
        pattern_exp_class[self.axis] = 0

        if self.target_is_sparse:  # convert to 1-hot probabilistic like coding
            classes = T.arange(self.n_class)
            classes = classes.dimshuffle(pattern_exp_class)
            if self.n_indep == 1:  # assuming target (b, ...)
                # target = target.dimshuffle(pattern_add_class)
                target = T.addbroadcast(target, self.axis)
                target = T.eq(target, classes)  # to 1-hot
            else:  # assuming target (b, n_indep, ...)
                t = []
                for i in range(self.n_indep):
                    component = target[:, i:i + 1]
                    component = T.addbroadcast(component, self.axis)
                    t.append(T.eq(component, classes))
                target = T.concatenate(t, axis=self.axis)

        # Target is now a 1-hot encoded bool of shape pred.shape

        if self.class_weights is None:
            class_weights = 1
        else:
            if isinstance(self.class_weights, Node):
                class_weights = self.class_weights.output
            else:
                class_weights = self.class_weights

            class_weights = class_weights.dimshuffle(pattern_exp_class)
            assert class_weights.ndim == pred.ndim

        if self.example_weights is None:
            example_weights = 1
        else:
            example_weights = self.example_weights.output
            example_weights = example_weights.dimshuffle(pattern_add_class)
            assert example_weights.ndim == pred.ndim
        if self.mask_class_labeled is not None:
            m_pattern = ['x', ] * pred.ndim
            m_pattern[self.axis] = 0
            m_pattern[self.pred.shape.tag2index('b')] = 1
            mask_class_labeled = self.mask_class_labeled.output.dimshuffle(
                m_pattern)

            target = target * mask_class_labeled  # this excludes some classes
            # in target (set their row to 0)

        nll_up = -xlogy0(target * class_weights * example_weights, pred + EPS)
        n_labelled_up = target.sum()

        if self.mask_class_not_present is not None:
            m_pattern = ['x', ] * pred.ndim
            m_pattern[self.axis] = 1
            m_pattern[self.pred.shape.tag2index('b')] = 0
            # Expand the mask to the full size, because below we want to sum it
            mask_class_not_present = self.mask_class_not_present.output. \
                                         dimshuffle(m_pattern) * T.ones_like(
                target)
            nll_dn = -xlogy0(
                mask_class_not_present * class_weights * example_weights,
                1.0 - pred + EPS)
            n_labelled_dn = mask_class_not_present.sum()

        else:
            nll_dn = 0.0
            n_labelled_dn = 0.0

        # Scale by n_labelled and n_indep
        # because the x-entropy is the sum across the classes, but this sum
        # is not taken here (so the pred.size is n_class times to big, when
        # given to AggregateLoss)
        n_tot = n_labelled_up + n_labelled_dn
        nll = (nll_up + nll_dn) * pred.size / (
        n_tot + EPS) / self.n_indep / self.n_class
        nll = T.sum(nll, axis=self.axis, keepdims=True)
        self._debug_outputs.extend([n_tot, pred.size])
        self.output = nll

    def _calc_shape(self):
        sh = self.parent[0].shape.updateshape(self.axis, 1)
        self.shape = sh


class BlockedMultinoulliNLL(Node):
    """
    Returns the symbolic mean and instance-wise negative log-likelihood of the prediction
    of this model under a given target distribution.

    Parameters
    ----------
    pred: Node
        Prediction node.
    target: T.Tensor
        corresponds to a vector that gives the correct label for each example. Labels < 0 are ignored (e.g. can
        be used for label propagation).
    blocking_factor: float
        Blocking factor.
    target_is_sparse: bool
        If the target is sparse.
    class_weights: T.Tensor
        weight vector of float32 of length  ``n_lab``. Values: ``1.0`` (default), ``w < 1.0`` (less important),
        ``w > 1.0`` (more important class).
    example_weights: T.Tensor
        weight vector of float32 of shape ``(bs, z, x, y)`` that can give the individual examples (i.e. labels for
        output pixels) different weights. Values: ``1.0`` (default), ``w < 1.0`` (less important),
        ``w > 1.0`` (more important example). Note: if this is not normalised/bounded it may result in a
        effectively modified learning rate!

    The following refers to lazy labels, the masks are always on a per patch basis, depending on the
    origin cube of the patch. The masks are properties of the individual image cubes and must be loaded
    into CNNData.

    mask_class_labeled: T.Tensor
        shape = (batchsize, num_classes).
        Binary masks indicating whether a class is properly labeled in ``y``. If a class ``k``
        is (in general) present in the image patches **and** ``mask_class_labeled[k]==1``, then
        the labels  **must** obey ``y==k`` for all pixels where the class is present.
        If a class ``k`` is present in the image, but was not labeled (-> cheaper labels), set
        ``mask_class_labeled[k]=0``. Then all pixels for which the ``y==k`` will be ignored.
        Alternative: set ``y=-1`` to ignore those pixels.
        Limit case: ``mask_class_labeled[:]==1`` will result in the ordinary NLL.
    mask_class_not_present: T.Tensor
        shape = (batchsize, num_classes).
        Binary mask indicating whether a class is present in the image patches.
        ``mask_class_not_present[k]==1`` means that the image does **not** contain examples of class ``k``.
        Then for all pixels in the patch, class ``k`` predictive probabilities are trained towards ``0``.
        Limit case: ``mask_class_not_present[:]==0`` will result in the ordinary NLL.
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.

    Examples
    --------
    - A cube contains no class ``k``. Instead of labelling the remaining classes they can be
      marked as unlabelled by the first mask (``mask_class_labeled[:]==0``, whether ``mask_class_labeled[k]``
      is ``0`` or ``1`` is actually indifferent because the labels should not be ``y==k`` anyway in this case).
      Additionally ``mask_class_not_present[k]==1`` (otherwise ``0``) to suppress predictions of ``k`` in
      in this patch. The actual value of the labels is indifferent, it can either be ``-1`` or it could be the
      background class, if the background is marked as unlabelled (i.e. then those labels are ignored).

    - Only part of the cube is densely labelled. Set ``mask_class_labeled[:]=1`` for all classes, but set the
      label values in the unlabelled part to ``-1`` to ignore this part.

    - Only a particular class ``k`` is labelled in the cube. Either set all other label pixels to ``-1`` or the
      corresponding flags in ``mask_class_labeled`` for the unlabelled classes.

    ..  Note::
        Using ``-1`` labels or telling that a class is not labelled, is somewhat redundant and just
        supported for convenience.
    """

    def __init__(self, pred, target, blocking_factor=0.5,
                 target_is_sparse=False, class_weights=None,
                 example_weights=None, mask_class_labeled=None,
                 mask_class_not_present=None, name="nll", print_repr=True):
        ###TODO add comment on normalisation
        parents = [pred, target]
        if class_weights is not None:
            parents.append(class_weights)
        if example_weights is not None:
            parents.append(example_weights)
        if mask_class_labeled is not None:
            parents.append(mask_class_labeled)
        if mask_class_not_present is not None:
            parents.append(mask_class_not_present)

        super(BlockedMultinoulliNLL, self).__init__(parents, name, print_repr)

        if isinstance(pred, Softmax):
            parent = pred
        else:
            if isinstance(pred, FromTensor) and isinstance(pred.parent,
                                                           Softmax):
                parent = pred.parent  # splitted softmax...
            else:
                raise ValueError(
                    "The prob input to a MultinoulliNLL-node must be "
                    "a Softmax-Node.")

        self.target = target
        self.pred = pred
        self.axis = pred.shape.tag2index('f')
        self.n_class = parent.n_class
        self.n_indep = parent.n_indep
        self.target_is_sparse = target_is_sparse
        self.class_weights = class_weights
        self.example_weights = example_weights
        self.mask_class_labeled = mask_class_labeled
        self.mask_class_not_present = mask_class_not_present
        self.blocking_factor = VariableParam(blocking_factor,
                                             name='blocking_factor',
                                             apply_train=False,
                                             apply_reg=False)

    def _make_output(self):
        """ Computation of Theano Output """
        pred = self.pred.output
        target = self.target.output

        pattern_add_class = list(range(pred.ndim - 1))
        pattern_add_class.insert(self.axis, 'x')

        pattern_exp_class = ['x', ] * pred.ndim
        pattern_exp_class[self.axis] = 0

        if self.target_is_sparse:  # convert to 1-hot probabilistic like coding
            classes = T.arange(self.n_class)
            classes = classes.dimshuffle(pattern_exp_class)
            if self.n_indep == 1:  # assuming target (b, ...)
                # target = target.dimshuffle(pattern_add_class)
                target = T.addbroadcast(target, self.axis)
                target = T.eq(target, classes)  # to 1-hot
            else:  # assuming target (b, n_indep, ...)
                t = []
                for i in range(self.n_indep):
                    component = target[:, i:i + 1]
                    component = T.addbroadcast(component, self.axis)
                    t.append(T.eq(component, classes))
                target = T.concatenate(t, axis=self.axis)

        # Target is now a 1-hot encoded bool of shape pred.shape

        if self.class_weights is None:
            class_weights = 1
        else:
            class_weights = self.class_weights.output
            class_weights = class_weights.dimshuffle(pattern_exp_class)
            assert class_weights.ndim == pred.ndim
        if self.example_weights is None:
            example_weights = 1
        else:
            example_weights = self.example_weights.output
            example_weights = example_weights.dimshuffle(pattern_add_class)
            assert example_weights.ndim == pred.ndim
        if self.mask_class_labeled is not None:
            m_pattern = ['x', ] * pred.ndim
            m_pattern[self.axis] = 0
            m_pattern[self.pred.shape.tag2index('b')] = 1
            mask_class_labeled = self.mask_class_labeled.output.dimshuffle(
                m_pattern)

            target = target * mask_class_labeled  # this excludes some classes
            # in target (set their row to 0)

        # Blocking
        b_pattern = [slice(None)] * pred.ndim
        b_pattern[self.axis] = slice(1, None)
        new_pred = T.maximum(
            self.blocking_factor * pred[b_pattern].max(axis=self.axis),
            pred[b_pattern])
        T.set_subtensor(pred[b_pattern], new_pred)

        nll_up = -xlogy0(target * class_weights * example_weights, pred + EPS)
        n_labelled_up = target.sum()

        if self.mask_class_not_present is not None:
            m_pattern = ['x', ] * pred.ndim
            m_pattern[self.axis] = 1
            m_pattern[self.pred.shape.tag2index('b')] = 0
            # Expand the mask to the full size, because below we want to sum it
            mask_class_not_present = self.mask_class_not_present.output. \
                                         dimshuffle(m_pattern) * T.ones_like(
                target)
            nll_dn = -xlogy0(
                mask_class_not_present * class_weights * example_weights,
                1.0 - pred + EPS)
            n_labelled_dn = mask_class_not_present.sum()

        else:
            nll_dn = 0.0
            n_labelled_dn = 0.0

        # Scale by n_labelled and n_indep
        # because the x-entropy is the sum across the classes, but this sum
        # is not taken here (so the pred.size is n_class times to big, when
        # given to AggregateLoss)
        n_tot = n_labelled_up + n_labelled_dn
        nll = (nll_up + nll_dn) * pred.size / (
        n_tot + EPS) / self.n_indep / self.n_class
        nll = T.sum(nll, axis=self.axis, keepdims=True)
        self._debug_outputs.extend([new_pred, n_tot, pred.size])
        self.output = nll

    def _calc_shape(self):
        sh = self.parent[0].shape.updateshape(self.axis, 1)
        self.shape = sh


class MalisNLL(Node):
    """
    Malis NLL node. (See https://github.com/TuragaLab/malis)

    Parameters
    ----------
    pred: Node
        Prediction node.
    aff_gt: T.Tensor
    seg_gt: T.Tensor
    nhood: np.ndarray
    unrestrict_neg: bool
    class_weights: T.Tensor
        weight vector of float32 of length  ``n_lab``. Values: ``1.0`` (default), ``w < 1.0`` (less important),
        ``w > 1.0`` (more important class).
    example_weights: T.Tensor
        weight vector of float32 of shape ``(bs, z, x, y)`` that can give the individual examples (i.e. labels for
        output pixels) different weights. Values: ``1.0`` (default), ``w < 1.0`` (less important),
        ``w > 1.0`` (more important example). Note: if this is not normalised/bounded it may result in a
        effectively modified learning rate!
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, pred, aff_gt, seg_gt, nhood, unrestrict_neg=True,
                 class_weights=None, example_weights=None,
                 name="nll", print_repr=True):
        parents = [pred, aff_gt, seg_gt]
        if class_weights is not None:
            parents.append(class_weights)
        if example_weights is not None:
            parents.append(example_weights)

        super(MalisNLL, self).__init__(parents, name, print_repr)

        if not isinstance(pred, Softmax):
            raise ValueError("The prob input to a MultinoulliNLL-node must be "
                             "a Softmax-Node.")
        if pred.shape['b'] != 1:
            raise NotImplementedError(
                "Malis can only be used with batch size 1.")

        self.aff_gt = aff_gt
        self.seg_gt = seg_gt
        self.pred = pred
        self.nhood = np.asarray(nhood, dtype=np.int32)
        self.unrestrict_neg = unrestrict_neg
        self.axis = pred.shape.tag2index('f')
        self.n_class = pred.n_class
        self.n_indep = pred.n_indep
        self.class_weights = class_weights
        self.example_weights = example_weights

    def _make_output(self):
        """ Computation of Theano Output """

        from ..malis.malisop import malis_weights

        pred = self.pred.output
        aff_gt = self.aff_gt.output[0]  # strip batch (1)
        seg_gt = self.seg_gt.output[0, 0]  # strip batch (1) and #class (1)

        pattern_add_class = list(range(pred.ndim - 1))
        pattern_add_class.insert(self.axis, 'x')

        pattern_exp_class = ['x', ] * pred.ndim
        pattern_exp_class[self.axis] = 0
        if self.class_weights is None:
            class_weights = 1
        else:
            class_weights = self.class_weights.output
            class_weights = class_weights.dimshuffle(pattern_exp_class)[
                0]  # strip batch dimension
            assert class_weights.ndim == pred[0].ndim
        if self.example_weights is None:
            example_weights = 1
        else:
            example_weights = self.example_weights.output
            example_weights = example_weights.dimshuffle(pattern_add_class)[
                0]  # strip batch dimension
            assert example_weights.ndim == pred[0].ndim

        sl = [slice(None), ] * pred.ndim
        sl[self.axis] = slice(1, None, self.n_class)
        # pred.shape = (bs, 6, x, y, z) 6--> edge1 neg, edge1 pos, edge2 neg...
        affinity_pred = pred[tuple(sl)][0]  # strip batch dimension

        sl = [slice(None), ] * pred.ndim
        sl[self.axis] = slice(0, None, self.n_class)
        disconnect_pred = pred[tuple(sl)][0]  # strip batch dimension

        pos_count, neg_count = malis_weights(affinity_pred,
                                             aff_gt,
                                             seg_gt,
                                             self.nhood,
                                             self.unrestrict_neg)

        pos_weight = pos_count * example_weights * class_weights
        neg_weight = neg_count * example_weights * class_weights
        weighted_pos = xlogy0(pos_weight,
                              affinity_pred + EPS)  # drive up prediction for "connected" here
        weighted_neg = xlogy0(neg_weight,
                              disconnect_pred + EPS)  # drive down prediction for "disconnected" here
        n_pos = T.sum(pos_count)
        n_neg = T.sum(neg_count)
        n_tot = n_pos + n_neg
        nll = -(weighted_pos + weighted_neg)
        # Scale by n_tot, because the counts n_tot are greater
        # than pred.size (~N**2), but the actual value depends on the amount
        # of ECS in the example
        self.output = nll * T.cast(nll.size, 'float32') / (n_tot + EPS)

        # For debug/inspection, take care that those are not in self.output
        false_splits = T.sum((affinity_pred < 0.5) * pos_count)
        false_merges = T.sum((affinity_pred > 0.5) * neg_count)
        rand_index = T.cast(false_splits + false_merges, 'float32') / (
        n_tot + EPS)
        self.rand_index = rand_index
        self.false_splits = false_splits
        self.false_merges = false_merges
        self.pos_count = pos_count
        self.neg_count = neg_count

        # eg 0.0   5187779 4578211 9765990 3439497477 7732598379 1143.9798583984375)
        # return nll, n_pos, n_neg, n_tot, false_splits, false_merges, rand_index, pos_count, neg_count

    def _calc_shape(self):
        sh = self.parent[0].shape.updateshape(self.axis, 1)
        self.shape = sh


class Classification(Node):
    """
    Classification node.

    Parameters
    ----------
    pred: Node
        Prediction node.
    n_class
    n_indep
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, pred, n_class='auto', n_indep='auto', name="cls",
                 print_repr=True):
        super(Classification, self).__init__(pred, name, print_repr)
        if not isinstance(pred, Softmax):
            if pred.activation_func in ['sig', 'logistic', 'sigmoid']:
                self.n_class = 2
                self.n_indep = pred.shape['f']
                self.sm_input = False
            else:
                assert n_class != 'auto'
                assert n_indep != 'auto'
                self.n_class = n_class
                self.n_indep = n_indep
                self.sm_input = n_indep != pred.shape['f']
        else:  # pred is softmax node
            self.n_class = pred.n_class
            self.n_indep = pred.n_indep
            self.sm_input = True

        self.pred = pred

    def _make_output(self):
        """ Computation of Theano Output """
        n_class = self.n_class
        n_indep = self.n_indep
        pred = self.pred.output
        axis = self.pred.shape.tag2index('f')
        if self.sm_input:
            if self.n_indep == 1:
                cls = T.argmax(pred, axis=axis, keepdims=True)

            else:
                y = []
                for i in range(n_indep):
                    sl = [slice(None), ] * pred.ndim
                    sl[axis] = slice(i * n_class, (i + 1) * n_class, 1)
                    cls = T.argmax(pred[tuple(sl)], axis=axis, keepdims=True)
                    y.append(cls)

                cls = T.concatenate(y, axis=axis)
        else:
            cls = T.gt(pred, 0.5)

        self.output = cls

    def _calc_shape(self):
        sh = self.parent.shape.updateshape(self.pred.shape.tag2index('f'),
                                           self.n_indep)
        self.shape = sh


class _Errors(Node):
    """
    Errors node.

    Parameters
    ----------
    cls: T.Tensor
    target: T.Tensor
        corresponds to a vector that gives the correct label for each example. Labels < 0 are ignored (e.g. can
        be used for label propagation).
    target_is_sparse: bool
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, cls, target, target_is_sparse=False,
                 name="errors", print_repr=True):
        parents = [cls, target]
        super(_Errors, self).__init__(parents, name, print_repr)

        self.n_class = cls.n_class
        self.n_indep = cls.n_indep

        self.target = target
        self.cls = cls
        self.target_is_sparse = target_is_sparse

    def _make_output(self):
        """ Computation of Theano Output """
        n_class = self.n_class
        n_indep = self.n_indep
        target = self.target.output
        axis = self.cls.shape.tag2index('f')

        if not self.target_is_sparse:
            if self.n_indep == 1:
                gt = T.argmax(target, axis=axis, keepdims=True)
            else:
                gt = []
                # This assumes that target is (b,n_class*n_indep,x,y,z)
                for i in range(n_indep):
                    sl = [slice(None), ] * target.ndim
                    sl[axis] = slice(i * n_class, (i + 1) * n_class, 1)
                    t = T.argmax(target[tuple(sl)], axis=axis, keepdims=True)
                    # t = T.argmax(target[:,i*n_class:(i+1)*n_class], axis=axis, keepdims=True)
                    gt.append(t)
                gt = T.concatenate(gt, axis=axis)
        else:
            gt = target

        gt = T.cast(gt, 'int16')

        self.output = T.mean(T.neq(gt, self.cls.output))

    def _calc_shape(self):
        self.shape = TaggedShape([1, ], ['f', ])


def Errors(pred, target, target_is_sparse=False, n_class='auto', n_indep='auto',
           name="errors", print_repr=True):
    if not isinstance(pred, Classification):
        pred = Classification(pred, n_class=n_class, n_indep=n_indep,
                              name='cls for errors', print_repr=False)
    return _Errors(pred, target, target_is_sparse=target_is_sparse,
                   name=name, print_repr=print_repr)


class GaussianNLL(Node):
    """
    Similar to squared loss but "modulated" in scale by the variance.

    Parameters
    ----------
    target: Node
        True value (target), usually directly an input node
    mu: Node
        Mean of the predictive Gaussian density
    sig: Node
        Sigma of the predictive Gaussian density
    sig_is_log: bool
        Whether ``sig`` is actaully the ln(sig), then it is
        exponentiated internally


    Computes element-wise:

     .. math::

       0.5 \cdot  ( ln(2  \pi \sigma)) + (target-\mu)^2/\sigma^2 )

    """

    def __init__(self, mu, sig, target, sig_is_log=False, name="g_nll",
                 print_repr=True):
        super(GaussianNLL, self).__init__((mu, sig, target), name, print_repr)

        self.target = target
        self.mu = mu
        self.sig = sig
        self.sig_is_log = sig_is_log

    def _make_output(self):
        """ Computation of Theano Output """
        target = self.target.output
        mu = self.mu.output
        sig = self.sig.output

        # IF there are several samples per instance the target must be made
        # broadcastable along the sample axis
        if 's' in self.mu.shape.tags:
            pattern = list(range(self.target.type.ndim))
            batch_index = self.mu.shape.tag2index('s')
            pattern.insert(batch_index, 'x')
            target = target.dimshuffle(pattern)

        if self.sig_is_log:
            log_sig = sig
            sig = T.exp(sig)
        else:
            log_sig = T.log(sig)

        normalisation = 0.5 * np.log(2 * np.pi) + log_sig
        gauss = 0.5 * ((target - mu) / sig) ** 2
        logpxz = normalisation + gauss

        self.output = logpxz


class BetaNLL(Node):
    """
    Similar to BinaryNLL loss but "modulated" in scale by the variance.

    Parameters
    ----------

    target: Node
        True value (target), usually directly an input node, must be in range [0,1]
    mode: Node
        Mode of the predictive Beta density, must come from linear
        activation function (will be transformed by exp(.) + 2 )
    concentration: node
        concentration of the predictive Beta density


    Computes element-wise:

     .. math::

       0.5 \cdot  2
    """

    def __init__(self, mode, concentration, target, name="beta_nll",
                 print_repr=True):
        super(BetaNLL, self).__init__((mode, concentration, target), name,
                                      print_repr)

        self.target = target
        self.mode = mode
        self.concentration = concentration

    def _make_output(self):
        """ Computation of Theano Output """
        target = self.target.output
        mode = self.mode.output
        concentration = self.concentration.output

        # IF there are several samples per instance the target must be made
        # broadcastable along the sample axis
        if 's' in self.mode.shape.tags:
            pattern = list(range(self.target.type.ndim))
            batch_index = self.mode.shape.tag2index('s')
            pattern.insert(batch_index, 'x')
            target = target.dimshuffle(pattern)

        def log_inv_beta_func(a, b):
            return T.gammaln(a + b) - T.gammaln(a) - T.gammaln(b)

        def log_beta_pdf(x, mode, concentration):
            a = mode * (concentration - 2) + 1
            b = (1 - mode) * (concentration - 2) + 1
            p = log_inv_beta_func(a, b) + (a - 1) * T.log(x + EPS) + (
                                                                     b - 1) * T.log(
                1 - x + EPS)
            return p

        concentration2 = concentration
        self.output = - log_beta_pdf(target, mode,
                                     concentration2) + T.nnet.softplus(
            -concentration)  # sign!!!


class BinaryNLL(Node):
    """
    Binary NLL node. Identical to cross entropy.

    Parameters
    ----------

    pred: Node
        Predictive Bernoulli probability.
    target: Node
        True value (target), usually directly an input node.


    Computes element-wise:

     .. math::

       -(target  ln(pred) + (1 - target) ln(1 - pred))
    """

    def __init__(self, pred, target, subtract_label_entropy=False,
                 name="binary_nll", print_repr=True):
        super(BinaryNLL, self).__init__((pred, target), name, print_repr)

        self.target = target
        self.pred = pred
        self.pred_shape = pred.shape
        self.subtract_label_entropy = subtract_label_entropy

    def _make_output(self):
        target = self.target.output
        pred = self.pred.output
        # IF there are several samples per instance the target must be made
        # broadcastable along the sample axis
        if 's' in self.pred_shape.tags:
            pattern = list(range(target.type.ndim))
            batch_index = self.pred_shape.tag2index('s')
            pattern.insert(batch_index, 'x')
            target = target.dimshuffle(pattern)

        # mask = T.isnan(self.target)
        mask = T.isclose(target, -666.0)
        logger.warning(
            "BinaryNLL: isnan is replaced by 'isclose(target, -666)'")
        n_labelled = (1 - mask).sum()
        n_tot = pred.size
        scale = T.cast(n_tot, 'float32') / (n_labelled + 1)

        # logpxz = T.nnet.binary_crossentropy(pred, target) # This makes NaNs!!!!
        logpxz = -xlogy0(target, pred + EPS) - xlogy0(1.0 - target,
                                                      1.0 - pred + EPS)
        if self.subtract_label_entropy:
            logpxz += -xlogy0(target, target + EPS) - xlogy0(1.0 - target,
                                                             1.0 - target + EPS)

        logpxz = T.set_subtensor(logpxz[mask.nonzero()], 0.0)
        logpxz *= scale
        self.output = logpxz
        self._debug_outputs.extend([n_tot, n_labelled, scale, pred, target])


class SquaredLoss(Node):
    """
    Squared loss node.

    Parameters
    ----------
    pred: Node
        Prediction node.
    target: T.Tensor
        corresponds to a vector that gives the correct label for each example. Labels < 0 are ignored (e.g. can
        be used for label propagation).
    margin: float or None
    scale_correction: float or None
        Downweights absolute deviations for large target scale. The value specifies
        the target value at which the square deviation has half weight compared to target=0
        If the target is twice as large as this value the downweight is 1/3 and so on.
        Note: the smaller this value the stronger the effect. No effect would be
        +inf
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, pred, target, margin=None, scale_correction=None,
                 name="se", print_repr=True):
        super(SquaredLoss, self).__init__((pred, target), name, print_repr)

        self.target = target
        self.pred = pred

        if margin:
            margin = VariableParam(value=margin, name="margin", dtype=floatX,
                                   apply_train=False)
            self.params['margin'] = margin

        self.margin = margin

        if scale_correction:
            scale_correction = VariableParam(value=scale_correction,
                                             name="scale_correction",
                                             dtype=floatX,
                                             apply_train=False)
            self.params['scale_correction'] = scale_correction

        self.scale_correction = scale_correction

    def _make_output(self):
        """ Computation of Theano Output """
        target = self.target.output
        pred = self.pred.output

        # IF there are several samples per instance the target must be made
        # broadcastable along the sample axis
        # if 's' in self.mu.shape.tags:
        #     pattern = list(range(self.target.type.ndim))
        #     batch_index = self.mu.shape.tag2index('s')
        #     pattern.insert(batch_index, 'x')
        #     target = target.dimshuffle(pattern)

        # mask = T.isnan(target)
        mask = T.isclose(target, -666.0)
        logger.warning(
            "SquaredLoss: isnan is replaced by 'isclose(target, -666)'")
        n_labelled = (1 - mask).sum()
        n_tot = pred.size
        scale = T.cast(n_tot, 'float32') / (n_labelled + 1)

        if self.margin is not None:
            diff = target - pred
            out = scale * 0.5 * T.square(diff) * T.ge(abs(diff),
                                                      self.margin) - self.margin
        else:
            out = scale * 0.5 * T.square(target - pred)

        if self.scale_correction is not None:
            correction = self.scale_correction / (
            abs(target) + self.scale_correction)
            out *= correction

        out = T.set_subtensor(out[mask.nonzero()], 0.0)
        self.output = T.mean(out, axis=self.pred.shape.tag2index('f'),
                             keepdims=True)
        self._debug_outputs.extend([n_tot, n_labelled, scale, pred, target])

    def _calc_shape(self):
        sh = self.parent[0].shape.updateshape(self.pred.shape.tag2index('f'), 1)
        self.shape = sh


class EuclideanDistance(Node):
    """
    Euclidian distance node.

    Parameters
    ----------
    pred: Node
        Prediction node.
    target: T.Tensor
        corresponds to a vector that gives the correct label for each example. Labels < 0 are ignored (e.g. can
        be used for label propagation).
    margin: float/None
    scale_correction: float/None
        Downweights absolute deviations for large target scale. The value specifies
        the target value at which the square deviation has half weight compared to target=0
        If the target is twice as large as this value the downweight is 1/3 and so on.
        Note: the smaller this value the stronger the effect. No effect would be
        +inf
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """  # TODO: Docstring Parameters do not match __init__ parameters.

    def __init__(self, pred, target, name="se", print_repr=True):
        super(EuclideanDistance, self).__init__((pred, target), name,
                                                print_repr)

        self.target = target
        self.pred = pred

    def _make_output(self):
        """ Computation of Theano Output """
        target = self.target.output
        pred = self.pred.output

        # IF there are several samples per instance the target must be made
        # broadcastable along the sample axis

        diff = target - pred
        out = diff.norm(2, axis=self.pred.shape.tag2index('f'))
        mask = T.isnan(out)
        self.output = T.set_subtensor(out[mask.nonzero()], 0.0)
        self._debug_outputs.extend([pred, target])

    def _calc_shape(self):
        sh = self.parent[0].shape.updateshape(self.pred.shape.tag2index('f'), 1)
        self.shape = sh


class RampLoss(Node):
    """
    RampLoss node.

    Parameters
    ----------
    pred: Node
        Prediction node.
    target: T.Tensor
        corresponds to a vector that gives the correct label for each example. Labels < 0 are ignored (e.g. can
        be used for label propagation).
    margin: float/None
    scale_correction: float/None
        downweights absolute deviations for large target scale. The value specifies
        the target value at which the square deviation has half weight compared to target=0
        If the target is twice as large as this value the downweight is 1/3 and so on.
        Note: the smaller this value the stronger the effect. No effect would be
        +inf
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """  # TODO: Docstring parameters do not match __init__ parameters.

    def __init__(self, d_low, d_big, name="se", print_repr=True, margin=None):
        super(RampLoss, self).__init__((d_low, d_big), name, print_repr)

        if margin is None:
            margin = 0
        margin = VariableParam(value=margin, name="margin", dtype=floatX,
                               apply_train=False)
        self.params['margin'] = margin

        self.margin = margin

        self.d_low = d_low
        self.d_big = d_big

    def _make_output(self):
        """ Computation of Theano Output """
        d_low = self.d_low.output
        d_big = self.d_big.output

        # IF there are several samples per instance the target must be made
        # broadcastable along the sample axis

        diff = d_low - d_big + self.margin
        neg_mask = diff < 0
        diff = T.set_subtensor(diff[neg_mask.nonzero()], 0.0)
        mask = T.isnan(diff)
        out = T.set_subtensor(diff[mask.nonzero()], 0.0)
        self.output = T.mean(out, axis=self.d_low.shape.tag2index('f'),
                             keepdims=True)
        self._debug_outputs.extend([d_low, d_big])

    def _calc_shape(self):
        sh = self.parent[0].shape.updateshape(self.d_low.shape.tag2index('f'),
                                              1)
        self.shape = sh


class AbsLoss(SquaredLoss):
    """
    AbsLoss node.

    Parameters
    ----------
    pred: Node
        Prediction node.
    target: T.Tensor
        corresponds to a vector that gives the correct label for each example. Labels < 0 are ignored (e.g. can
        be used for label propagation).
    margin: float or None
    scale_correction: float or None
        Boosts loss for large target values: if target=1 the error
        is multiplied by this value (and linearliy for other targets)
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, pred, target, margin=None, scale_correction=None,
                 name="absloss", print_repr=True):
        super(AbsLoss, self).__init__(pred, target, margin=margin,
                                      scale_correction=scale_correction,
                                      name=name, print_repr=print_repr)

    def _make_output(self):
        """ Computation of Theano Output """
        target = self.target.output
        pred = self.pred.output

        # IF there are several samples per instance the target must be made
        # broadcastable along the sample axis
        # if 's' in self.mu.shape.tags:
        #     pattern = list(range(self.target.type.ndim))
        #     batch_index = self.mu.shape.tag2index('s')
        #     pattern.insert(batch_index, 'x')
        #     target = target.dimshuffle(pattern)

        # mask = T.isnan(target)
        mask = T.isclose(target, -666.0)
        logger.warning(
            "SquaredLoss: isnan is replaced by 'isclose(target, -666)'")
        n_labelled = (1 - mask).sum()
        n_tot = pred.size
        scale = T.cast(n_tot, 'float32') / (n_labelled + 1)

        if self.margin is not None:
            diff = target - pred
            out = scale * abs(diff) * T.ge(abs(diff), self.margin) - self.margin
        else:
            out = scale * abs(target - pred)

        if self.scale_correction is not None:
            correction = self.scale_correction * abs(target) + 1.0
            out *= correction

        out = T.set_subtensor(out[mask.nonzero()], 0.0)
        self.output = T.mean(out, axis=self.pred.shape.tag2index('f'),
                             keepdims=True)
        self._debug_outputs.extend([n_tot, n_labelled, scale, pred, target])


class AggregateLoss(Node):
    """
    This node is used to average the individual losses over a batch
    (and possibly, spatial/temporal dimensions). Several losses can be
    mixed for multi-target training.

    Parameters
    ----------

    parent_nodes: list/tuple of graph or single node
        each component is some (possibly element-wise) loss array
    mixing_weights: list/None
        Weights for the individual costs. If none, then all are weighted
        equally. If mixing weights are used, they can be changed during
        training by manipulating the attribute ``params['mixing_weights']``.
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.

    # The following is all wrong, mixing_weights are directly used:

    The losses are first summed per component, and then the component sums
    are summed using the relative weights. The resulting scalar is finally
    normalised such that:
        * The cost does not grow with the number of mixed components
        * Components which consist of more individual losses have more weight
          e.g. If there is a constraint on some hidden representation
          with 20 features and a constraint the reconstruction of 100 features,
          the reconstruction constraint has 5x more impact on the overall loss
          than the constraint on the hidden state (provided those two loss
          are initially on the same scale). If they are intended to have equal
          impact, the weights should be used to upscale the constraint against
          the reconstruction.

    """  # TODO: What about the "all wrong" section?

    def __init__(self, parent_nodes, mixing_weights=None, name="total_loss",
                 print_repr=True):

        if not isinstance(parent_nodes, (tuple, list)):
            parent_nodes = [parent_nodes, ]

        super(AggregateLoss, self).__init__(parent_nodes, name, print_repr)

        if mixing_weights is None:
            mixing_weights = np.ones(len(parent_nodes))

        if isinstance(mixing_weights, (tuple, list, np.ndarray)):
            if len(parent_nodes) != len(mixing_weights):
                stat = "len(parent_nodes)=%i, len(weights)=%i" \
                       % (len(parent_nodes), len(mixing_weights))
                raise ValueError("Mismatch: %s" % (stat,))

            mixing_weights = np.array(mixing_weights, dtype=floatX)
            # mixing_weights *= (len(mixing_weights) / mixing_weights.sum()) # normalise
            mixing_weights = VariableParam(value=mixing_weights,
                                           name="loss_mixing_weights",
                                           dtype=floatX,
                                           apply_train=False)

        else:
            raise ValueError("Unsupported weight format")

        self.params['mixing_weights'] = mixing_weights
        self.mixing_weights = mixing_weights

    def _make_output(self):
        """ Computation of Theano Output """
        # The normalisation is such that:
        # - The cost does not grow with the number of mixed components
        # - Components which consist of more individual losses have more weight
        # inputs = [inp.output for inp in self.parent]
        # sums   = T.stack([inp.sum() for inp in inputs])
        # total_sum = T.sum(sums * self.mixing_weights)
        # sizes = [inp.size for inp in inputs]
        # total_size = T.sum(T.stack(sizes))
        # self.output = total_sum / total_size
        means = []
        for inp in self.parent:
            m = T.mean(inp.output)
            means.append(m)

        means = T.mul(means, self.mixing_weights)
        self.output = T.mean(means)

    def _calc_shape(self):
        self.shape = TaggedShape([1, ], ['f', ])

    def _calc_comp_cost(self):
        self.computational_cost = np.sum(
            [inp.shape.stripnone_prod for inp in self.parent])


def SobelizedLoss(pred, target, loss_type='abs', loss_kwargs=None):
    """
    SobelizedLoss node.

    Parameters
    ----------
    pred: Node
            Prediction node.
    target: T.Tensor
        corresponds to a vector that gives the correct label for each example. Labels < 0 are ignored (e.g. can
        be used for label propagation).
    loss_type: str
        Only "abs" is supported.
    loss_kwargs: dict
        kwargs for the AbsLoss constructor.
    Returns
    -------
    Node:
        The loss node.
    """
    if loss_kwargs is None:
        loss_kwargs = dict()

    dim = pred.shape.ndim
    f = pred.shape['f']

    w_sh = (f * dim, f) + (3,) * dim
    w = np.zeros(w_sh, dtype=floatX)
    b = np.zeros((f * dim), dtype=floatX)
    base_w = np.array([1, 0, -1], dtype=floatX)

    if dim > 1:
        n = np.array([[0.3, 0.4, 0.3]], dtype=np.float32).T
        base_w = np.tile(base_w, [3, 1]) * n
        base_w = np.concatenate([base_w[None], base_w.T[None]], axis=0)

    if dim > 2:
        base_w = np.tile(base_w[0], [3, 1, 1]) * n[:, :, None]
        base_w = np.concatenate([np.transpose(base_w, (2, 1, 0))[None],
                                 base_w[None],
                                 np.transpose(base_w, (1, 2, 0))[None]], axis=0)

    if dim > 3:
        raise NotImplementedError()

    # Now w_base has a filter for each dimension, next we need to take care
    # of the channels in the input
    for i in range(f):
        w[i::f, i] = base_w

    pred_sobel = Conv(pred, f * dim, (3,) * dim, (1,) * dim, conv_mode='same',
                      activation_func='lin', w=[w, 'const'], b=[b, 'const'],
                      name='pred_sobel')

    target_sobel = Conv(target, f * dim, (3,) * dim, (1,) * dim,
                        conv_mode='same',
                        activation_func='lin', w=[w, 'const'], b=[b, 'const'],
                        name='target_sobel')

    if loss_type == 'abs':
        loss = AbsLoss(pred_sobel, target_sobel, **loss_kwargs)
    # elif loss_type=='mnll':
    #    loss = MultinoulliNLL(pred_sobel, target_sobel, **loss_kwargs)
    else:
        raise NotImplementedError()

    return loss


if __name__ == "__main__":
    from elektronn2.neuromancer import Input

    #    pred = Input((2,6,1), 'b,f,x')
    #    pred = Softmax(pred, 3, 2)
    #    lab  = Input((2,2,1), 'b,f,x', name='labels', dtype='int16')
    #    cls  = Classification(pred)
    #    err  = Errors(pred, lab, target_is_sparse=True)
    #
    #    pred_val = np.array([[0.6,0.2,0.2,0.8,0.1,0.1],
    #                         [0.2,0.7,0.1,0.1,0.1,0.8]], dtype=np.float32)[...,None]
    #    lab_val = np.array([[0,0],[1,2]], dtype=np.int16)[...,None]
    #    print cls(pred_val), cls(pred_val).shape
    #    print err(pred_val, lab_val)
    #
    #
    #    lab  = Input((2,6,1), 'b,f,x', name='labels', dtype='int16')
    #    cls  = Classification(pred)
    #    err  = Errors(pred, lab, target_is_sparse=False)
    #
    #    pred_val = np.array([[0.6,0.2,0.2,0.8,0.1,0.1],
    #                         [0.2,0.7,0.1,0.1,0.1,0.8]], dtype=np.float32)[...,None]
    #    lab_val = np.array([[1,0,0,1,0,0],[0,1,0,0,0,1]], dtype=np.int16)[...,None]
    #    print cls(pred_val), cls(pred_val).shape
    #    print err(pred_val, lab_val)


    #    pred = Input((2,6), 'b,f')
    #    pred = Softmax(pred, 3, 2)
    #    lab  = Input((2,2), 'b,f', name='labels', dtype='int16')
    #    nll  = MultinoulliNLL(pred, lab, target_is_sparse=True)
    #
    #    pred_val = np.array([[0.6,0.2,0.2,0.8,0.1,0.1],
    #                         [0.2,0.7,0.1,0.1,0.1,0.8]], dtype=np.float32)
    #    lab_val = np.array([[0,0],[1,2]], dtype=np.int16)
    #    print nll(pred_val, lab_val), pred(pred_val)


    pred = Input((2, 6), 'b,f')
    example_weights = Input((2,), 'b', name='example_weights')
    class_weights = Input((2,), 'b', name='class_weights')
    mask_class_not_present = Input((2, 6), 'b,f', name='mask_class_not_present')
    lab = Input((2, 2), 'b,f', name='labels', dtype='int16')
    pred = Softmax(pred, 3, 2)

    nll = MultinoulliNLL(pred, lab,
                         example_weights=example_weights,
                         class_weights=class_weights,
                         mask_class_not_present=mask_class_not_present,
                         target_is_sparse=True)

    pred_val = np.array([[0.6, 0.2, 0.2, 0.8, 0.1, 0.1],
                         [0.2, 0.7, 0.1, 0.1, 0.1, 0.8]], dtype=np.float32)
    lab_val = np.array([[0, 0], [1, 2]], dtype=np.int16)
    exp_val = np.array([1, 1], dtype=np.int16)
    cls_val = np.array([1, 1, 1, 1, 1, 1], dtype=np.float32)
    not_pres = np.array([[0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0]],
                        dtype=np.int16)
    logger.debug(nll(pred_val, lab_val, cls_val, exp_val, not_pres))



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved
from __future__ import absolute_import

from .cnndata import *
from .traindata import *
#from .skeleton import *



# -*- coding: utf-8 -*-
# ELEKTRONN2 - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, \
    super, zip

import os
from collections import OrderedDict

from matplotlib import pyplot as plt
from scipy import stats
import numpy as np
import seaborn as sns

from .locking import FileLock

import logging


logger = logging.getLogger('elektronn2log')


class Scroller(object):
    def __init__(self, axes, images, names, init_z=None):
        self.axes = axes
        for ax in axes:
            ax.grid(b=False)
        # ax.set_title('use scroll wheel to navigate images')

        self.images = list(map(np.ascontiguousarray, images))
        self.n_slices = images[0].shape[0]
        self.ind = self.n_slices // 2 if init_z is None else init_z

        self.imgs = []
        for ax, dat, name in zip(axes, images, names):
            if name in ['id', 'ids', 'ID', 'IDs', 'seg', 'SEG', 'Seg', 'lab',
                        'label', 'Label']:
                cmap = 'nipy_spectral'
            else:
                cmap = 'gray'

            self.imgs.append(
                ax.imshow(dat[self.ind], interpolation='None', cmap=cmap))
            ax.set_xlabel(name)

        self.update()

    def onscroll(self, event):
        # print ("%s %s" % (event.button, event.step))
        if event.button=='up':
            self.ind = np.clip(self.ind + 1, 0, self.n_slices - 1)
        else:
            self.ind = np.clip(self.ind - 1, 0, self.n_slices - 1)

        self.update()

    def update(self):
        for ax, im, dat in zip(self.axes, self.imgs, self.images):
            im.set_data(dat[self.ind])
            ax.set_ylabel('slice %s' % self.ind)
            im.axes.figure.canvas.draw()


def scroll_plot(images, names=None, init_z=None):
    """
    Creates a plot 1x2 image plot of 3d volume images
    Scrolling changes the displayed slices

    Parameters
    ----------

    images: list of arrays (or single)
      Each array of shape (z,y,x) or (z,y,x,RGB)
    names: list of strings (or single)
      Names for each image

    Usage
    -----

    For the scroll interation to work, the "scroller" object
    must be returned to the calling scope

    >>> fig, scroller = _scroll_plot4(images, names)
    >>> fig.show()

    """
    if names is None:
        n = 1 if isinstance(images, np.ndarray) else len(images)
        names = [str(i) for i in range(n)]

    if isinstance(images, np.ndarray):
        return _scroll_plot1(images, names, init_z)
    elif len(images)==2:
        assert len(names) >= 2
        return _scroll_plot2(images, names, init_z)
    elif len(images)==4:
        assert len(names) >= 4
        return _scroll_plot4(images, names, init_z)


def _scroll_plot1(image, name, init_z):
    """
    Creates a plot of 3d volume images
    Scrolling changes the displayed slices

    Parameters
    ----------

    images:  array of shape (z,x,y) or (z,x,y,RGB)

    Usage
    -----

    For the scroll interation to work, the "scroller" object
    must be returned to the calling scope

    >>> fig, scroller = scroll_plot(image, name)
    >>> fig.show()

    """
    fig = plt.figure(figsize=(12, 12))
    ax1 = fig.add_subplot(111)

    scroller = Scroller([ax1], [image, ], [name, ], init_z)
    fig.canvas.mpl_connect('scroll_event', scroller.onscroll)
    fig.tight_layout()
    return scroller


def _scroll_plot2(images, names, init_z):
    """
    Creates a plot 1x2 image plot of 3d volume images
    Scrolling changes the displayed slices

    Parameters
    ----------

    images: list of 2 arrays
      Each array of shape (z,y,x) or (z,y,x,RGB)
    names: list of 2 strings
      Names for each image

    Usage
    -----

    For the scroll interation to work, the "scroller" object
    must be returned to the calling scope

    >>> fig, scroller = _scroll_plot4(images, names)
    >>> fig.show()

    """
    fig = plt.figure(figsize=(12, 6))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122, sharex=ax1, sharey=ax1)

    scroller = Scroller([ax1, ax2], images, names, init_z)
    fig.canvas.mpl_connect('scroll_event', scroller.onscroll)
    fig.tight_layout()
    return scroller


def _scroll_plot4(images, names, init_z):
    """
    Creates a plot 2x2 image plot of 3d volume images
    Scrolling changes the displayed slices

    Parameters
    ----------

    images: list of 4 arrays
      Each array of shape (z,y,x) or (z,y,x,RGB)
    names: list of 4 strings
      Names for each image

    Usage
    -----

    For the scroll interation to work, the "scroller" object
    must be returned to the calling scope

    >>> fig, scroller = _scroll_plot4(images, names)
    >>> fig.show()

    """
    fig = plt.figure(figsize=(12, 12))
    ax1 = fig.add_subplot(221)
    ax2 = fig.add_subplot(222, sharex=ax1, sharey=ax1)
    ax3 = fig.add_subplot(223, sharex=ax1, sharey=ax1)
    ax4 = fig.add_subplot(224, sharex=ax1, sharey=ax1)

    scroller = Scroller([ax1, ax2, ax3, ax4], images, names, init_z)
    fig.canvas.mpl_connect('scroll_event', scroller.onscroll)
    fig.tight_layout()
    return scroller


def _embed3d2d(a, border_width=1, normalize=False, output_ratio=1.5, ):
    """
    Embed an 3d array into an 2d matrix by tiling.
    The last two dimensions of ``a`` are assumed to be spatial, the first is tiled.
    """
    sh = a.shape
    assert len(sh)==3
    n = sh[0]
    nhor = int(np.ceil(np.sqrt(n * output_ratio)))  # aim: ratio 16:9
    nvert = int(np.ceil(float(n) / nhor))  # warning: too big: nvert*nhor >= n

    if normalize:
        maxs = [np.max(a[i, :, :]) + 1e-8 for i in range(n)]
        mins = [np.min(a[i, :, :]) for i in range(n)]
    else:
        maxs = [1] * n
        mins = [0] * n

    ret = np.zeros(
        (nvert * (border_width + sh[1]), nhor * (border_width + sh[2])),
        dtype=np.float32)

    for j in range(nvert):
        for i in range(nhor):
            if i + j * nhor >= n:
                return ret
            ret[j*(border_width+sh[1]):j*(border_width+sh[1])+sh[1],
                i*(border_width+sh[2]):i*(border_width+sh[2])+sh[2]] = \
               (a[i+j*nhor,:,:]-mins[i+j*nhor])/(maxs[i+j*nhor]-mins[i+j*nhor])

    return ret


def embedfilters(filters, border_width=1, normalize=False, output_ratio=1.0,
                 rgb_axis=None):
    """
    Embed an nd array into an 2d matrix by tiling.
    The last two dimensions of ``a`` are assumed to be spatial,
    the others are tiled recursively.
    """
    if rgb_axis is not None:
        assert filters[rgb_axis]==3
        channels = []
        for i in range(3):
            slice = [slice(None), ] * filters.ndim
            slice[rgb_axis] = i
            f = filters[slice]
            channels.append(
                embedfilters(f, border_width, normalize, output_ratio))
        return np.dstack(channels)

    if filters.ndim==3:
        return _embed3d2d(filters, border_width, normalize, output_ratio)
    elif filters.ndim > 3:
        parts = []
        for f in filters:
            parts.append(
                embedfilters(f, border_width, normalize, output_ratio))
        parts = np.concatenate([x[None, ...] for x in parts])
        return embedfilters(parts, border_width, normalize, output_ratio)


def sma(c, n):
    """
    Returns box-SMA of c with box length n, the returned array has the same
    length as c and is const-padded at the beginning
    """
    if n==0:
        return c
    ret = np.cumsum(c, dtype=float)
    ret[n:] = (ret[n:] - ret[:-n]) / n
    m = min(n, len(c))
    ret[:n] = ret[:n] / np.arange(1, m + 1)  # unsmoothed
    return ret


def add_timeticks(ax, times, steps, time_str='mins', num=5):
    N = int(times[-1])
    k = max(N / num, 1)
    k = int(np.log10(k))  # 10-base of locators
    m = int(np.round(float(N) / (num * 10 ** k)))  # multiple of base
    s = max(m * 10 ** k, 1)
    x_labs = np.arange(0, N, s, dtype=np.int)
    x_ticks = np.interp(x_labs, times, steps)
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(x_labs)
    ax.set_xlim(0, steps[-1])
    ax.set_xlabel('Runtime [%s]' % time_str)  # (%s)'%("{0:,d}".format(N)))


def plot_hist(timeline, history, save_name, loss_smoothing_length=200,
              autoscale=True):
    """Plot graphical info during Training"""
    plt.ioff()
    try:
        # Subsample points for plotting
        N = len(timeline)
        x_timeline = np.arange(N)
        s = max((len(timeline) // 2000), 1)
        x_timeline = x_timeline[::s]
        timeline = timeline[::s]
        s = max((len(history) // 2000), 1)
        history = history[::s]

        if timeline['time'][-1] < 120 * 60:
            runtime = str(int(timeline['time'][-1] / 60)) + ' mins'
        else:
            runtime = "%.1f hours" % (timeline['time'][-1] / 3600)

        # check if valid data is available
        if not np.any(np.isnan(history['valid_loss'])):
            l = history['valid_loss'][-10:]
        else:
            l = history['train_loss'][-10:]

        loss_cap = l.mean() + 2 * l.std()
        lt = timeline['loss'][-200:]
        lt_m = lt.mean()
        lt_s = lt.std()
        loss_cap_t = lt_m + 2 * lt_s
        loss_cap = np.maximum(loss_cap, loss_cap_t)

        if np.all(timeline['loss'] > 0):
            loss_floor = 0.0
        else:
            loss_floor = lt_m - 2 * lt_s

        ### Timeline, Loss ###
        plt.figure(figsize=(16, 12))
        plt.subplot(211)
        plt.plot(x_timeline, timeline['loss'], 'b-', alpha=0.5,
                 label='Update Loss')
        loss_smooth = sma(timeline['loss'], loss_smoothing_length)
        plt.plot(x_timeline, loss_smooth, 'k-', label='Smooth update Loss',
                 linewidth=3)

        if autoscale:
            plt.ylim(loss_floor, loss_cap)
            plt.xlim(0, N)

        plt.legend(loc=0)
        plt.hlines(lt_m, 0, N, linestyle='dashed', colors='r', linewidth=2)
        plt.hlines(lt_m + lt_s, 0, N, linestyle='dotted', colors='r',
                   linewidth=1)
        plt.hlines(lt_m - lt_s, 0, N, linestyle='dotted', colors='r',
                   linewidth=1)
        plt.xlabel('Update steps %s, total runtime %s' % (N - 1, runtime))

        ax = plt.twiny()
        if timeline['time'][-1] > 120 * 60:
            add_timeticks(ax, timeline['time'] / 3600, x_timeline,
                          time_str='hours')
        else:
            add_timeticks(ax, timeline['time'] / 60, x_timeline,
                          time_str='mins')

        ### Loss vs Prevalence ###
        plt.subplot(212)
        c = 1.0 - (timeline['time'] / timeline['time'].max())
        plt.scatter(timeline['batch_char'], timeline['loss'], c=c, marker='.',
                    s=80, cmap='gray', edgecolors='face')
        if autoscale:
            bc = timeline['batch_char'][-200:]
            bc_m = bc.mean()
            bc_s = bc.std()
            bc_cap = bc_m + 2 * bc_s
            if np.all(bc > 0):
                bc_floor = -0.01
            else:
                bc_floor = bc_m - 2 * bc_s
            plt.ylim(loss_floor, loss_cap)
            plt.xlim(bc_floor, bc_cap)

        plt.xlabel('Mean target of batch')
        plt.ylabel('Loss')

        plt.tight_layout()
        with FileLock('plotting'):
            plt.savefig(save_name + ".timeline.png", bbox_inches='tight')

        ###################################################################
        ### History Loss ###
        plt.figure(figsize=(16, 12))
        plt.subplot(311)
        plt.plot(history['steps'], history['train_loss'], 'g-',
                 label='Train Loss', linewidth=3)
        plt.plot(history['steps'], history['valid_loss'], 'r-',
                 label='Valid Loss', linewidth=3)
        if autoscale:
            plt.ylim(loss_floor, loss_cap)
            plt.xlim(0, history['steps'][-1])

        plt.legend(loc=0)
        # plt.xlabel('Update steps %s, total runtime %s'%(N-1, runtime))

        ax = plt.twiny()
        if timeline['time'][-1] > 120 * 60:
            add_timeticks(ax, timeline['time'] / 3600, x_timeline,
                          time_str='hours')
        else:
            add_timeticks(ax, timeline['time'] / 60, x_timeline,
                          time_str='mins')

        ### History Loss gains ###
        plt.subplot(312)
        plt.plot(history['steps'], history['loss_gain'], 'b-',
                 label='Loss Gain at update', linewidth=3)
        plt.hlines(0, 0, history['steps'][-1], linestyles='dotted')
        plt.plot(history['steps'], history['lr'], 'r-', label='LR',
                 linewidth=3)
        # plt.xlabel('Update steps %s, total runtime %s'%(N-1, runtime))
        plt.legend(loc=3)
        std = history['loss_gain'][:5].std() * 2 if len(history) > 6 else 1.0
        if autoscale:
            # add epsilon to suppress matplotlib warning in case of CG
            plt.ylim(-std, std + 1e-10)
            plt.xlim(0, history['steps'][-1])

        ax2 = plt.twinx()
        ax2.plot(history['steps'], history['mom'], 'r-', label='MOM')
        ax2.plot(history['steps'], history['gradnetrate'], 'r-',
                 label='GradNetRate')

        ax2.set_ylim(-1, 1)
        if autoscale:
            ax2.set_xlim(0, history['steps'][-1])
        ax2.legend(loc=4)

        ### Errors ###
        plt.subplot(313)
        cutoff = 2
        if len(history) > (cutoff + 1):
            history = history[cutoff:]
        # check if valid data is available
        if not np.any(np.isnan(history['valid_err'])):
            e = history['valid_err'][-200:]
        else:
            e = history['train_err'][-200:]
        e_m = e.mean()
        e_s = e.std()
        err_cap = e_m + 2 * e_s
        if np.all(e > 0):
            err_floor = 0.0
        else:
            err_floor = e_m - 2 * e_s

        plt.plot(history['steps'], history['train_err'], 'g--',
                 label='Train error', linewidth=1)
        plt.plot(history['steps'], history['valid_err'], 'r--',
                 label='Valid Error', linewidth=1)

        plt.plot(history['steps'], sma(history['train_err'], 8), 'g-',
                 label='Smooth train error', linewidth=3)
        if not np.any(np.isnan(sma(history['valid_err'], 8))):
            plt.plot(history['steps'], sma(history['valid_err'], 8), 'r-',
                     label='Smooth valid Error', linewidth=3)
        if autoscale:
            plt.ylim(err_floor, err_cap)
            plt.xlim(0, history['steps'][-1])

        plt.grid()
        plt.legend(loc=0)
        plt.xlabel('Update steps %s, total runtime %s' % (N - 1, runtime))
        plt.tight_layout()
        with FileLock('plotting'):
            plt.savefig(save_name + ".history.png", bbox_inches='tight')

    except ValueError:
        # When arrays are empty
        logger.warning("An error occurred during plotting.")


def plot_var(var, save_name):
    # [i, nll, nll.std, conc.mean, conc.std,]
    plt.figure(figsize=(16, 12))
    plt.subplot(211)
    plt.plot(var[:, 0], var[:, 1], 'b-', alpha=0.6)
    plt.plot(var[:, 0], sma(var[:, 1], 100), 'g-', linewidth=3)
    plt.plot(var[:, 0], sma(var[:, 1] + var[:, 2], 100), 'r:', linewidth=2)
    plt.plot(var[:, 0], sma(var[:, 1] - var[:, 2], 100), 'r:', linewidth=2)
    plt.title("NLL")

    plt.subplot(212)
    plt.plot(var[:, 0], var[:, 3], 'b-', alpha=0.6)
    plt.plot(var[:, 0], sma(var[:, 3], 100), 'g-', linewidth=3)
    plt.plot(var[:, 0], sma(var[:, 3] + var[:, 4], 100), 'r:', linewidth=2)
    plt.plot(var[:, 0], sma(var[:, 3] - var[:, 4], 100), 'r:', linewidth=2)
    plt.title("Concentration")

    with FileLock('plotting'):
        plt.savefig(save_name + ".Beta1.png", bbox_inches='tight')

    plt.figure(figsize=(12, 12))
    c = 1.0 - ((var[:, 0]).astype(np.float32) / var[-1, 0])

    plt.subplot(221)
    plt.scatter(var[:, 1], var[:, 3], c=c, marker='.', s=80, cmap='gray',
                edgecolors='face')
    plt.title("Concentration vs. NLL")

    plt.subplot(222)
    plt.scatter(var[:, 2], var[:, 3], c=c, marker='.', s=80, cmap='gray',
                edgecolors='face')
    plt.title("Concentration vs. NLL.std")

    plt.subplot(223)
    plt.scatter(var[:, 3], var[:, 4], c=c, marker='.', s=80, cmap='gray',
                edgecolors='face')
    plt.title("Concentration vs. Concentration.std")

    plt.subplot(224)
    plt.scatter(var[:, 1], var[:, 2], c=c, marker='.', s=80, cmap='gray',
                edgecolors='face')
    plt.title("NLL vs. NLL.std")

    with FileLock('plotting'):
        plt.savefig(save_name + ".Beta2.png", bbox_inches='tight')


def plot_debug(var, debug_output_names, save_name):
    # [i, nll, other....]
    s = max((len(var) // 2000), 1)
    var = var[::s]
    plt.figure(figsize=(16, 12))
    colors = ['gold', 'b', 'darkblue', 'crimson', 'navajowhite', 'deepskyblue',
              'darkgray', 'maroon', 'palevioletred', 'forestgreen', ] * 2
    n = len(colors) // 2
    marker = ['-', ] * n + [':'] * n
    lw_s = [2, ] * n + [3, ] * n
    maxima = []
    minima = []
    total = sma(var[:, 1], 70)
    maxima.append(total[-100:].max())
    minima.append(total[-100:].min())
    plt.plot(var[:, 0], total, 'k-', linewidth=4, label='total loss')
    for i in range(len(debug_output_names)):  ###TODO automatic std intervals
        name = debug_output_names[i]
        smooth = sma(var[:, i + 2], 70)
        plt.plot(var[:, 0], smooth, color=colors[i], linestyle=marker[i],
                 linewidth=lw_s[i], label=name)
        maxima.append(smooth[-100:].max())
        minima.append(smooth[-100:].min())

    plt.title("Debug Outputs")
    cap_hi = np.max([x for x in maxima if np.isfinite(x)]) * 1.5
    cap_lo = np.min([x for x in minima if np.isfinite(x)])
    plt.ylim(cap_lo, cap_hi)
    plt.legend(loc=0)
    plt.hlines(0, var[0, 0], var[-1, 0], linewidth=1)
    plt.grid()

    with FileLock('plotting'):
        plt.savefig(save_name + ".Debug.png", bbox_inches='tight')


def plot_regression(pred, target, save_name, loss_smoothing_length=200,
                    autoscale=True):
    """Plot graphical info during Training"""
    try:
        # Subsample points for plotting
        N = len(pred)
        s = max((len(pred) // 2000), 1)
        pred = pred[::s].ravel()
        target = target[::s].ravel()
        N = len(pred)
        x_timeline = np.arange(N)
        c = N - x_timeline
        plt.figure(figsize=(8, 8))
        ### Loss vs Prevalence ###
        plt.scatter(pred, target, c=c, marker='.', s=80, cmap='gray',
                    edgecolors='face')
        m = np.minimum(pred.min(), target.min())
        M = np.maximum(pred.max(), target.max())
        plt.plot([m, M], [m, M], 'r:')
        plt.ylim(m, M)
        plt.xlim(m, M)
        plt.xlabel('Prediction')
        plt.ylabel('Target')
        plt.tight_layout()
        with FileLock('plotting'):
            plt.savefig(save_name + ".regression.png", bbox_inches='tight')
    except ValueError:
        # When arrays are empty
        logger.warning("An error occurred during regression plotting.")


def plot_kde(pred, target, save_name, limit=90, scale='same', grid=50,
             take_last=4000):
    try:
        if take_last:
            pred = pred[-take_last:].ravel()
            target = target[-take_last:].ravel()

        if limit=='max':
            mp, mt = pred.min(), target.min()
            Mp, Mt = pred.max(), target.max()
        else:
            lo = 100 - limit
            mp, mt = np.percentile(pred, lo), np.percentile(target, lo)
            Mp, Mt = np.percentile(pred, limit), np.percentile(target, limit)

        if scale=='same':
            mp = np.minimum(mp, mt)
            Mp = np.maximum(Mp, Mt)
            mt = mp
            Mt = Mp

        if isinstance(grid, int):
            grid = [grid, grid]

        pg, tg = np.mgrid[mp:Mp:grid[0] * 1j, mt:Mp:grid[1] * 1j]

        positions = np.vstack([pg.ravel(), tg.ravel()])
        values = np.vstack([pred, target])
        kernel = stats.gaussian_kde(values)
        f = np.reshape(kernel(positions).T, pg.shape)
        plt.figure()
        plt.xlim(mp, Mp)
        plt.ylim(mt, Mt)
        plt.xlabel("Prediction")
        plt.ylabel("Target")
        plt.imshow(np.rot90(f), cmap=plt.cm.gist_earth_r,
                   extent=[mp, Mp, mt, Mt])
        plt.contour(pg, tg, f)
        plt.plot([mt, Mt], [mt, Mt], 'r:')
        plt.tight_layout()
        with FileLock('plotting'):
            plt.savefig(save_name + ".regression_kde.png", bbox_inches='tight')
    except ValueError:
        # When arrays are empty
        logger.warning("An error occurred during regression kde plotting.")


def my_quiver(x, y, img=None, c=None):
    """
    first dim of x,y changes along vertical axis
    second dim changes along horizontal axis
    x: vertical vector component
    y: horizontal vector component
    """
    figure = plt.figure(figsize=(7, 7))

    if img is not None:
        plt.imshow(img, interpolation='none', alpha=0.22, cmap='gray')

    plt.quiver(x, y, c, angles='xy', units='xy', cmap='spring', pivot='middle',
               scale=0.5)
    return figure


def plot_trainingtarget(img, lab, stride=1):
    """
    Plots raw image vs target to check if valid batches are produced.
    Raw data is also shown overlaid with targets

    Parameters
    ----------

    img: 2d array
      raw image from batch
    lab: 2d array
      targets
    stride: int
      strides of targets

    """

    if len(lab) * stride!=len(img):
        off = (len(img) - stride * len(lab)) // 2 // stride
        if lab.ndim==3:
            assert lab.shape[2]==3
            new_t = np.zeros(
                (lab.shape[0] + 2 * off, lab.shape[1] + 2 * off, 3))
            new_t[off:-off, off:-off, :] = lab
        else:
            new_t = np.zeros((lab.shape[0] + 2 * off, lab.shape[1] + 2 * off))
            new_t[off:-off, off:-off] = lab

        lab = new_t

    if lab.ndim==3:
        assert lab.shape[2]==3
        img = img[:, :, None]
        img = np.repeat(img, 3, axis=2)

    plt.figure(figsize=(18, 6))
    plt.subplot(131)
    plt.imshow(img, interpolation='none', cmap=plt.get_cmap('gray'))
    plt.title('data')
    plt.subplot(132)
    plt.imshow(lab, interpolation='none', cmap=plt.get_cmap('gray'))
    plt.title('target')
    if img.shape==lab.shape:
        overlay = 0.75 * img + 0.25 * (1 - lab)
        plt.subplot(133)
        plt.imshow(overlay, interpolation='none', cmap=plt.get_cmap('gray'))
        plt.title('overlay')
    plt.show()
    return img - lab


def plot_exectimes(exectimes, save_path='~/exectimes.png', max_items=32):
    """
    Plot model execution time dict obtained from
    elektronn2.neuromancer.model.Model.measure_exectimes()

    :param exectimes: OrderedDict of execution times
                      (output of Model.measure_exectimes())
    :param save_path: Where to save the plot
    :param max_items: Only the max_items largest execution times are given
                      names and are plotted independently.
                      Everything else is grouped under '(other nodes)'.
    """
    thresh_val = 0
    if len(exectimes) > max_items:
        thresh_val = sorted(list(exectimes.values()))[-max_items]
    filt_rtimes = OrderedDict()
    for key, val in exectimes.items():
        if val >= thresh_val:
            filt_rtimes[key] = val
    other = sum(exectimes.values()) - sum(filt_rtimes.values())
    node_names = list(filt_rtimes.keys())
    node_exectimes = list(filt_rtimes.values())
    if len(exectimes) > max_items:
        node_names += ['(other nodes)']
        node_exectimes += [other]
    cs = plt.cm.Set1(np.arange(len(node_exectimes)) / (len(node_exectimes)))

    sns.set_style("whitegrid")
    plt.figure(figsize=(13, 12))
    plt.title('Node execution times')
    plt.ylabel('Node')
    plt.xlabel('Time (in ms)')
    ax = sns.barplot(y=node_names, x=node_exectimes)
    with FileLock('plotting'):
        ax.get_figure().savefig(os.path.expanduser(save_path), bbox_inches='tight')



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

# TODO: Python 3 compatibility


__all__ = ['KnossosArray', 'KnossosArrayMulti']

import glob
import itertools
import re
import os
import sys
import traceback
import ctypes
import multiprocessing as mp
from collections import deque

import numpy as np

from .. import utils

if sys.version_info[:2] != (2, 7):
    raise ImportError(
        '\nSorry, this module only supports Python 2.7.'
        '\nYour current Python version is {}\n'.format(sys.version)
    )


def our_glob(s):
    l = []
    for g in glob.glob(s):
        l.append(g.replace(os.path.sep,"/"))
    return l

def get_remaining_size(fd):
    try:
        fn = fd.fileno()
    except AttributeError:
        return os.path.getsize(fd.name) - fd.tell()
    st = os.fstat(fn)
    size = st.st_size - fd.tell()
    return size

def load_rawfile(path, out, offset=0):
    fd = open(path, 'rb')
    if (offset > 0):
        fd.seek(offset, 1)
    size = get_remaining_size(fd)
    itemsize = out.itemsize
    shapeprod = out.size
    nbytes = shapeprod * itemsize
    if nbytes > size:
        raise ValueError(
                "Not enough bytes left in file for specified shape and type")

    nbytesread = fd.readinto(out.data)
    if nbytesread != nbytes:
        raise IOError("Didn't read as many bytes as expected")

    fd.close()
    return out

class LoadProc(mp.Process):
    def __init__(self, mp_array, path, dtype, status):
        super(LoadProc, self).__init__()
        self.path = path
        self.mp_array = mp_array
        self.status = status
        self.array = np.frombuffer(mp_array.get_obj(), dtype)

    def run(self):
        try:
            self.mp_array.acquire()
            a_buffer = np.empty(len(self.array), np.uint8)
            load_rawfile(self.path, a_buffer)
            self.array[:] = a_buffer
            self.array /= 255
            self.mp_array.release()
        except:
            self.mp_array.release()
            traceback.print_exc()
            self.status.set()


class KnossosArray(object):
    """
    Interfaces with knossos cubes, all  axes are in zxy order!
    """
    def __init__(self, path, max_ram=1000, n_preload=2, fixed_mag=1):
        path = os.path.expanduser(path)
        # Knossos attributes
        self._shape = np.zeros(3, dtype=np.int)
        self._shape_spatial = np.zeros(3, dtype=np.int)
        self.scale = np.zeros(3, dtype=np.float)
        self._knossos_path = None
        self._experiment_name = None
        self._cube_boundary = None
        self._cube_length = 128
        self._cube_sh = [self._cube_length,]*3
        self._mag = []

        self._n_f = 1

        # slicing/loading attributes
        self.dtype = np.dtype(np.float32)
        _n_preload = np.ones(3, dtype=np.float)*n_preload
        _n_preload[0] /= 2
        self.n_preload = np.ceil(_n_preload).astype(np.int)

        max_bytes = max_ram * 2**20
        self.cache_size = int(max_bytes / (self._cube_length**3 * self.dtype.itemsize))
        if self.cache_size<np.prod(self.n_preload*2+1):
            ram = np.prod(self.n_preload*2+1) * (self._cube_length**3 * self.dtype.itemsize) // 2**20
            raise ValueError("For %s preload cubes more RAM is required: %i MB" %(n_preload, ram))

        self.mpa_used = {}
        self.mpa_empty = []
        self.cubes = {}
        self.load_q = deque()
        self.loading = {}

        for i in range(self.cache_size):
            mp_array = mp.Array(ctypes.c_float, int(self._cube_length**3), lock=True)
            mp_array.acquire()
            self.mpa_empty.append(mp_array)
        self._initialize_from_knossos_path(path, fixed_mag)
        # self.track = []

    @property
    def shape(self):
        return self._shape

    @property
    def n_f(self):
        return self._n_f

    def __repr__(self):
        s = "<KnossosArray> %s\n"%(self.shape,)
        s += "%i cubes in cache\n" %len(self.cubes)
        s += "%i mpa in use\n" %len(self.mpa_used)
        s += "%i mpa empty\n" %len(self.mpa_empty)
        s += "%i mpa loading\n" %len(self.loading)
        s += "%i loading q\n" %len(self.load_q)
        return s

    def _initialize_from_knossos_path(self, path, fixed_mag=0):
        """ Initializes the dataset by parsing the knossos.conf in path + "mag1"

        :param path: str
            forward-slash separated path to the datasetfolder - not .../mag !
        :param fixed_mag: int
            fixes available mag to one specific value

        :return:
            nothing
        """
        self._knossos_path = path
        all_mag_folders = our_glob(os.path.join(path,"*mag*"))

        if fixed_mag > 0:
            self._mag.append(fixed_mag)
        else:
            for mag_test_nb in range(32):
                for mag_folder in all_mag_folders:
                    if "mag"+str(2**mag_test_nb) in mag_folder:
                        self._mag.append(2**mag_test_nb)
                        break

        mag_folder = our_glob(os.path.join(path,"*mag*"))[0].split("/")
        if len(mag_folder[-1]) > 1:
            mag_folder = mag_folder[-1]
        else:
            mag_folder = mag_folder[-2]

        self._name_mag_folder = \
            mag_folder[:-len(re.findall("[\d]+", mag_folder)[-1])]

        try:
            p = our_glob(os.path.join(path,"*mag1"))[0]+"/knossos.conf"
            print("Reading %s" %p)
            f = open(p)

            lines = f.readlines()
            f.close()
        except:
            raise NotImplementedError("Could not find/read *mag1/knossos.conf")

        parsed_dict = {}
        for line in lines:
            try:
                match = re.search(r'(?P<key>[A-Za-z _]+)'
                                  r'((?P<numeric_value>[0-9\.]+)'
                                  r'|"(?P<string_value>[A-Za-z0-9._/-]+)");',
                                  line).groupdict()

                if match['string_value']:
                    val = match['string_value']
                elif '.' in match['numeric_value']:
                    val = float(match['numeric_value'])
                elif match['numeric_value']:
                    val = int(match['numeric_value'])
                else:
                    raise Exception('Malformed knossos.conf')

                parsed_dict[match["key"]] = val
            except:
                print("Unreadable line in knossos.conf - ignored.")

        self._shape[2] = parsed_dict['boundary x ']
        self._shape[1] = parsed_dict['boundary y ']
        self._shape[0] = parsed_dict['boundary z ']
        self._shape_spatial[:] = self._shape
        self._cube_boundary = np.ceil(self._shape.astype(np.float) / self._cube_length).astype(np.int) # the boundary is the first for which no cube is availbale!
        self.scale[2] = parsed_dict['scale x ']
        self.scale[1] = parsed_dict['scale y ']
        self.scale[0] = parsed_dict['scale z ']

        self._experiment_name = parsed_dict['experiment name ']
        if self._experiment_name.endswith("mag1"):
            self._experiment_name = self._experiment_name[:-5]

    def _request_cube(self, coords):
        assert (coords not in self.cubes) and (coords not in self.loading)
        assert len(self.mpa_empty)
        # self.track.append(coords)
        coords_xyz = [coords[2], coords[1], coords[0]] # knossos has xyz order, we have zyx
        mag = self._mag[0]
        path = self._knossos_path+self._name_mag_folder+\
               str(mag) + "/x%04d/y%04d/z%04d/" \
               % (coords_xyz[0], coords_xyz[1], coords_xyz[2]) + \
               self._experiment_name + '_mag' + str(mag) + \
               "_x%04d_y%04d_z%04d.raw" \
               % (coords_xyz[0], coords_xyz[1], coords_xyz[2])
        if not os.path.exists(path):
            raise ValueError("Slice out of bounce, no cube %s "
            "[Note: not all cubes exist within bounding box!].)" %(coords,))

        mp_array = self.mpa_empty.pop()
        status = mp.Event() # is not set
        mp_array.release()
        proc = LoadProc(mp_array, path, self.dtype, status)
        proc.start()
        self.load_q.append([status, proc, mp_array, coords])
        self.loading[coords]=True

    def _clear_q(self, wait_for):
        if wait_for is None:
            states = [q[1].is_alive() for q in self.load_q]
            try:
                max_i = states.index(True)
            except: # all have terminated already
                max_i = len(self.load_q)

        elif wait_for=='all':
            max_i = len(self.load_q)
        else:
            max_i = len(self.load_q)

        i = 0
        while i<max_i:
            status, proc, mp_array, coords = self.load_q.popleft()
            if proc.is_alive():
                print("WAITING for cube %s" %(coords,))
            proc.join()

            if status.is_set():
                raise Exception("An error happened in one of the loading processes")

            mp_array.acquire()
            np_array = np.frombuffer(mp_array.get_obj(), self.dtype).reshape(self._cube_sh)
            assert coords not in self.cubes, "%s"%(coords,)
            self.cubes[coords] = np_array
            self.mpa_used[coords] = mp_array
            self.loading.pop(coords)
            #print("cube %s loaded" %(coords,))
            if coords==wait_for:
                break
            i += 1

    def preload(self, position, start_end=None, sync=False):
        """
        preloads around position preload distance but at least to cover start-end
        """
        position = np.array(position, dtype=np.int)
        center_cube = np.floor((position)/self._cube_length).astype(np.int)
        start = np.maximum(center_cube - self.n_preload, 0)
        end   = np.minimum(center_cube + self.n_preload + 1, self._cube_boundary)
        # preload indices
        ijk_pre = list(itertools.product(range(start[0], end[0]),
                                         range(start[1], end[1]),
                                         range(start[2], end[2])))

        # requested indices:
        if start_end is not None:
            start_r, end_r = start_end
            if np.any(end_r > self._cube_boundary):
                print("Warning: Slice at %s might be out of bounce (no cube %s)" %(position, end_r-1))

            ijk_req = list(itertools.product(range(start_r[0], end_r[0]),
                                             range(start_r[1], end_r[1]),
                                             range(start_r[2], end_r[2])))
            ijk_req_set = set((ijk_req))
            ijk_pre = list(filter(lambda x: x not in ijk_req_set, ijk_pre))
            ijk = ijk_req + ijk_pre
        else:
            ijk = ijk_pre

        fetch = []
        keep  = {}
        for coords in ijk:
            if (coords in self.cubes):
                keep[coords] = True
            elif (coords not in self.loading):
                fetch.append(coords)

        if len(fetch) > len(self.mpa_empty): # need to release arrays first
            n = len(fetch) - len(self.mpa_empty)
            if len(ijk) > self.cache_size:
                f = float(len(ijk)) / self.cache_size
                raise ValueError("The requested array+preload margin is %.2f x"
                " too large for the cache limit" %f)
            self._release_cubes(center_cube, keep, n)

        for i,j,k in fetch:
            self._request_cube((i,j,k))

        if sync:
            self._clear_q("all")

        if start_end is not None:
            return ijk_req

    def _release_cube(self, coords):
        self.cubes.pop(coords)
        mp_array = self.mpa_used.pop(coords)
        self.mpa_empty.append(mp_array)
        #print("cube %s deleted" %(coords,))

    def _release_cubes(self, center_cube, keep, n):
        release = []
        for coords in self.cubes:
            if coords not in keep:
                dist = np.linalg.norm((center_cube - coords) * [2,1,1])
                release.append((coords, dist))

        release.sort(key=lambda x: x[1], reverse=True)

        if len(release)< n: # not enough cubes available, they are still in loading q
            for i in range(len(release)): # release the possible cubes first
                self._release_cube(release[i][0])

            n_left = n - len(release)
            # now clear  the q
            try: # without waiting
                self._clear_q(None)
                self._release_cubes(center_cube, keep, n_left)
            except: # if it doesnt work, wait for all
                self._clear_q('all')
                self._release_cubes(center_cube, keep, n_left)
            return


        for i in range(n):
            self._release_cube(release[i][0])


    def _fetch_cube(self, coords):
        """
        Fetch a cube that is in cache or in loading q. i.e. this cube must have been preloaded
        Don't use this to request/preload a cube!
        """
        if coords not in self.cubes:
            assert coords in self.loading
            self._clear_q(coords)

        return self.cubes[coords]

    #@utils.timeit
    def cut_slice(self, shape, offset, out=None):
        #tt = utils.Timer()
        shape = np.array(shape, dtype=np.int)
        offset = np.array(offset, dtype=np.int)
        start = np.floor((offset)/self._cube_length).astype(np.int)
        end   = np.floor((offset+shape-1)/self._cube_length).astype(np.int) + 1
        position = offset + shape//2

        #tt.check("setup", True)
        ijk = self.preload(position=position, start_end=(start, end), sync=False)
        #tt.check("preload", True)
        if out is None:
            out = np.empty(shape, self.dtype)
        else:
            assert out.dtype==self.dtype
            assert np.allclose(out.shape, shape)

        l = self._cube_length
        for i,j,k in ijk: # Flat triple-nested loop
            values = self._fetch_cube((i,j,k))
            #tt.check("fetch", True)
            xl = max(offset[0]-i*l, 0)
            xh = min(offset[0]-i*l+shape[0], l)
            yl = max(offset[1]-j*l, 0)
            yh = min(offset[1]-j*l+shape[1], l)
            zl = max(offset[2]-k*l, 0)
            zh = min(offset[2]-k*l+shape[2], l)
            cut = values[xl:xh, yl:yh, zl:zh]
            #tt.check("cut", True)
            xl1 = max(i*l-offset[0], 0)
            xh1 = min((i+1)*l-offset[0], shape[0])
            yl1 = max(j*l-offset[1], 0)
            yh1 = min((j+1)*l-offset[1], shape[1])
            zl1 = max(k*l-offset[2], 0)
            zh1 = min((k+1)*l-offset[2], shape[2])
            out[xl1:xh1, yl1:yh1, zl1:zh1] = cut
            #tt.check('write', True)

        #tt.summary()
        return out

    def _normalise_idx(self, idx):
        new_idx = []
        shape = np.zeros(3, dtype=np.int)
        offset = np.zeros(3, dtype=np.int)
        for i,sl in enumerate(idx):
            if isinstance(sl, int):
                raise ValueError("Cannot slice single slices, only ranges")
            start, stop, step = sl.indices(self._shape_spatial[i])
            if start < 0 or stop>=self._shape_spatial[i]:
                raise ValueError("Slice %s out of bounce for knossos "
                                 "array of shape %s" %(idx[i], self._shape_spatial[i]))
            if step != 1:
                raise NotImplementedError("Subsampled slicing not implemented:"
                                          "instructins: use mag(stride) cubes "
                                          "and reuse slicing function")
            new_idx.append(slice(start, stop, step))
            shape[i] = (stop - start)//step
            offset[i] = start

        return shape, offset, new_idx

    def __getitem__(self, idx):
        assert len(idx) in [3,4]
        add_channel = False
        if len(idx)==4:
            assert idx[0].start==idx[0].stop==idx[0].step==None
            idx = idx[1:]
            add_channel = True
        shape, offset, new_idx = self._normalise_idx(idx)
        out = self.cut_slice(shape, offset)
        if add_channel:
            out = out[None]
        return out


class KnossosArrayMulti(KnossosArray):
    def __init__(self, path_prefix, feature_paths, max_ram=3000, n_preload=2, fixed_mag=1):
        self._arrays = []
        self._n_f = len(feature_paths)
        self.dtype = np.dtype(np.float32)
        for feature in feature_paths:
            p = os.path.expanduser(os.path.join(path_prefix,feature))
            self._arrays.append(KnossosArray(p, max_ram // self._n_f, n_preload, fixed_mag))


        self._shape = (self._n_f,) + tuple(self._arrays[0].shape)
        self._shape_spatial = self._arrays[0].shape

    def preload(self, position, sync=True):
        if len(position)==4:
            position = position[1:]

        for a in self._arrays:
            a.preload(position, sync=sync)

    def __getitem__(self, idx):
        assert len(idx)==4
        feature_slice = idx[0]
        idx = idx[1:]
        shape, offset, new_idx = self._normalise_idx(idx)
        feature_indices = list(range(*feature_slice.indices(self._n_f)))
        out = np.empty((len(feature_indices),)+tuple(shape), self.dtype)
        for i in feature_indices:
            self._arrays[i].cut_slice(shape, offset, out[i])

        return out

    def cut_slice(self, shape, offset, out=None):
        if out is None:
            out = np.empty((self._n_f,)+tuple(shape), self.dtype)

        for i in range(self._n_f):
            self._arrays[i].cut_slice(shape, offset, out[i])

        return out

    def __repr__(self):
        s = "<KnossosArray> %s\nFirst sub-array:\n"%(self.shape,)
        s += repr(self._arrays[0])
        return s


class KnossosArrayMultiMy(KnossosArrayMulti):
    def __getitem__(self, idx):
        tmp = super(KnossosArrayMultiMy, self).__getitem__(idx)
        sh = list(tmp.shape)
        sh[0] = 2
        out = np.empty(sh, dtype=np.float32)
        out[0] = tmp[0]
        out[1] = np.maximum(np.maximum(tmp[1], tmp[2]) - tmp[3], 0)
        return out

    @property
    def n_f(self):
        return self.n_f

    @property
    def shape(self):
        sh = list(self._shape)
        sh[0] = 2
        return tuple(sh)



if __name__ == "__main__":
    import time, sys
    sys.path.append(os.path.expanduser("~/axon/mkilling/devel/"))
    from knossos_utils import KnossosDataset
    from elektronn2.utils.plotting import scroll_plot, scroll_plot2
    path = os.path.expanduser("~/lustre/sdorkenw/j0126_3d_rrbarrier/")
    arr = KnossosArray(path, max_ram=1200, n_preload=2)

    arr.preload([0,0,0], sync=True)
    sh = np.array([120,240,240])
    off = np.array([0,0,0]) + np.random.randint(0, 500, size=3)

#    a = arr[0:80, 0:220, 0:230]
#    a0 = arr[0:100, 0:400, 0:200]
#    a1 = arr[0:100, 100:400+100, 100:200+100]
#    b = arr[0:640, 0:640, 0:384]
#    c = arr[500:600, 500:800, 500:800]
#
    kds = KnossosDataset()
    kds.initialize_from_knossos_path(path)
    get = utils.timeit(kds.from_raw_cubes_to_matrix)

#    a0 = arr.cut_slice(sh, off)
#    a1 = get(sh[[2,1,0]], off[[2,1,0]])
#    a2 = (a1.T).astype(np.float32)/255
#    print(np.allclose(a0, a2))
#    fig = scroll_plot2([a0, a2], 'ab')

    t = 0
    out =  np.empty(sh, dtype=np.float32)
    N = 500
    for i in range(N):
        t0 = time.time()
        a = arr.cut_slice(sh, off, out=out)
        t += time.time()-t0

        time.sleep(0.2)
        off  += np.random.randint(0, 10, size=3)


    print("%.3f slices/s" %(float(N)/t))

    arr._clear_q(None)
    fig = scroll_plot(out, 'a')










